from typing import Any, List, Iterator

from prefect import Client
from prefect.engine.state import Scheduled, State
from prefect.engine.result import Result, NoResultType
from prefect.utilities.graphql import with_args, EnumValue
from prefect.utilities.logging import get_logger

# Utility to indicate a task result is not loaded to distinguish from `None` values
NotLoaded = object()

logger = get_logger("backend.flow_run")


class TaskRunView:
    """
    A view of Task Run data stored in the Prefect API.

    Provides lazy loading of task run return values from Prefect `Result` locations.

    This object is designed to be an immutable view of the data stored in the Prefect
    backend API at the time it is created.

    Args:
        - task_run_id: The task run uuid
        - task_id: The uuid of the task associated with this task run
        - task_slug: The slug of the task associated with this task run
        - name: The task run name
        - state: The state of the task run
        - map_index: The map index of the task run. Is -1 if it is not a mapped subtask,
             otherwise it is in the index of the task run in the mapping
        - flow_run_id: The uuid of the flow run associated with this task run
    """

    def __init__(
        self,
        task_run_id: str,
        task_id: str,
        task_slug: str,
        name: str,
        state: State,
        map_index: int,
        flow_run_id: str,
    ) -> None:
        self.task_run_id = task_run_id
        self.name = name
        self.task_id = task_id
        self.task_slug = task_slug
        self.state = state
        self.map_index = map_index
        self.flow_run_id = flow_run_id

        # Uses NotLoaded so to separate from return values of `None`
        self._result: Any = NotLoaded

    def get_result(self) -> Any:
        """
        The result of this task run loaded from the `Result` location. Lazily loaded
        on the first call then cached for repeated access. For the parent of mapped
        task runs, this will include the results of all children. May require
        credentials to be present if the result location is remote (ie S3). If your
        flow was run on another machine and `LocalResult` was used, we will fail
        to load the result.

        See `TaskRunView.iter_mapped` for lazily iterating through mapped tasks instead
        of retrieving all of the results at once.

        Returns:
            Any: The value your task returned
        """

        if not self.state.is_finished():
            raise ValueError("The task result cannot be loaded if it is not finished.")

        self._assert_result_type_is_okay()

        if self._result is NotLoaded:
            # Load the result from the result location
            self._result = self._load_result()

        return self._result

    def _load_result(self) -> Any:
        if self.state.is_mapped():
            self._load_child_results()

        # Fire the state result hydration
        self.state.load_result()
        return self.state.result

    def _load_child_results(self) -> None:
        """
        Mapped tasks require `state.map_states` to be manually filled with the states
        of child tasks for the result to be hydrated correctly

        If the parent flow run is not `Finished`, we'll run into some interesting
        problems during this retrieval as the mapped children can be uncreated, have
        null states, or not be finished yet. We attempt to raise helpful errors in this
        case, but it's not really a supported paradigm.
        """
        if not self.state.is_mapped():
            raise ValueError("Child results cannot be loaded for an unmapped task.")

        # Load all the child task runs
        child_task_runs = [
            self._from_task_run_data(task_run)
            for task_run in self._query_for_task_runs(
                where={
                    "task": {"slug": {"_eq": self.task_slug}},
                    "flow_run_id": {"_eq": self.flow_run_id},
                    # Ignore the root task since we are the root task
                    "map_index": {"_neq": -1},
                },
                # Ensure the returned tasks are ordered matching map indices
                order_by={"map_index": EnumValue("asc")},
                # Handle errors cleanly below
                error_on_empty=False,
            )
        ]

        # Raise an informative error if none were found
        if not child_task_runs:
            raise ValueError(
                f"No child task runs were found for task {self.task_slug!r} in "
                f"flow run {self.flow_run_id}. Is the flow run finished?"
            )

        # Ensure that all the children are finished
        for task_run in child_task_runs:
            if not task_run.state.is_finished():
                raise ValueError(
                    f"Child task run {task_run.task_slug}[{task_run.map_index}] is "
                    f"in state {task_run.state}. The result cannot be retrieved. "
                    "Loading results for mapped tasks in running flows is not "
                    "supported."
                )

            # Ensure the mapped children have valid result types
            task_run._assert_result_type_is_okay()

            # Update state
            self.state.map_states = [task_run.state for task_run in child_task_runs]

    def _assert_result_type_is_okay(self) -> None:
        """
        Since we do not have access to the user's custom Result class, we cannot load
        the result.
        """
        # Mapped parents do not have result types
        if self.state.is_mapped():
            return

        # If there is no result type, we cannot load it. Mapped parents don't apply
        if not self.state._result or type(self.state._result) is NoResultType:
            raise TypeError(
                "The task has a no `Result` type so the result cannot be loaded."
                "Set a `Result` type on your tasks so return values are persisted."
            )

        # Must have a location to be loaded
        if not self.state._result.location:
            raise ValueError(
                "The task result has no `location` so the result cannot be loaded. "
                "This often means that your task result has not been configured or has "
                "been configured incorrectly."
            )

        # TODO: Check for custom serializer types as well once they are added to the
        #       `ResultSchema`
        # TODO: Add the `result_type` to `TaskRunView` so we can report the qualified
        #       name of the custom result type used.
        if type(self.state._result) is Result:
            # The `Result` base class is used during deserialization if a custom result
            # class has been declared unless it's a mapped state in which it will be
            # an empty `Result` and this assertion will be called on the map children
            raise TypeError(
                "The task has a custom `Result` type and its result cannot be loaded. "
                "Only built-in `Result` types are supported."
            )

    def iter_mapped(self) -> Iterator["TaskRunView"]:
        """
        Iterate over the results of a mapped task, yielding a `TaskRunView` for each map
        index. This query is not performed in bulk so the results can be lazily
        consumed. If you want all of the task results at once, use `result` instead.

        Yields:
            A `TaskRunView` for each mapped item
        """
        if not self.state.is_mapped():
            raise TypeError(
                f"Task run {self.task_run_id!r} ({self.task_slug}) is not a "
                "mapped task."
            )

        # Generate a where clause given the map index
        where = lambda index: {
            "task": {"slug": {"_eq": self.task_slug}},
            "flow_run_id": {"_eq": self.flow_run_id},
            "map_index": {"_eq": index},
        }
        map_index = 0
        while True:  # Iterate until we are out of child task runs
            task_run_data = self._query_for_task_run(
                where=where(map_index), error_on_empty=False
            )
            if not task_run_data:
                break

            yield self._from_task_run_data(task_run_data)
            map_index += 1

    @classmethod
    def _from_task_run_data(cls, task_run: dict) -> "TaskRunView":
        """
        Instantiate a `TaskRunView` from serialized data

        This method deserializes objects into their Prefect types.

        Args:
            - task_run: The serialized task run data

        Returns:
            A populated `TaskRunView` instance
        """
        task_run = task_run.copy()  # Create a copy to avoid mutation
        task_run_id = task_run.pop("id")
        task_data = task_run.pop("task")

        # The serialized state _could_ be null if the backend has not
        # created it yet, this would typically be seen with mapped tasks
        serialized_state = task_run.pop("serialized_state") or Scheduled().serialize()

        return cls(
            task_run_id=task_run_id,
            state=State.deserialize(serialized_state),
            task_id=task_data["id"],
            task_slug=task_data["slug"],
            **task_run,
        )

    @classmethod
    def from_task_run_id(cls, task_run_id: str = None) -> "TaskRunView":
        """
        Get an instance of this class; query by task run id

        Args:
            - task_run_id: The UUID identifying the task run in the backend

        Returns:
            A populated `TaskRunView` instance
        """
        if not isinstance(task_run_id, str):
            raise TypeError(
                f"Unexpected type {type(task_run_id)!r} for `task_run_id`, "
                f"expected 'str'."
            )

        return cls._from_task_run_data(
            cls._query_for_task_run(where={"id": {"_eq": task_run_id}})
        )

    @classmethod
    def from_task_slug(
        cls, task_slug: str, flow_run_id: str, map_index: int = -1
    ) -> "TaskRunView":
        """
        Get an instance of this class; query by task slug and flow run id.

        Args:
            - task_slug: The unique string identifying this task in the flow. Typically
                `<task-name>-1`.
            - flow_run_id: The UUID identifying the flow run the task run occurred in
            - map_index (optional): The index to access for mapped tasks; defaults to
                the parent task with a map index of -1

        Returns:
            A populated `TaskRunView` instance
        """
        return cls._from_task_run_data(
            cls._query_for_task_run(
                where={
                    "task": {"slug": {"_eq": task_slug}},
                    "flow_run_id": {"_eq": flow_run_id},
                    "map_index": {"_eq": map_index},
                }
            )
        )

    @staticmethod
    def _query_for_task_run(where: dict, **kwargs: Any) -> dict:
        """
        Query for task run data using `_query_for_task_runs` but throw an exception if
        more than one matching task run is found

        Args:
            - where: The `where` clause to use
            - **kwargs: Additional kwargs are passed to `_query_for_task_runs`

        Returns:
            A dict of task run data
        """
        task_runs = TaskRunView._query_for_task_runs(where=where, **kwargs)

        if len(task_runs) > 1:
            raise ValueError(
                f"Found multiple ({len(task_runs)}) task runs while querying for task "
                f"runs where {where}: {task_runs}"
            )

        if not task_runs:
            # Erroring on an empty result is handled by `_query_for_task_runs`
            return {}

        task_run = task_runs[0]
        return task_run

    @staticmethod
    def _query_for_task_runs(
        where: dict,
        order_by: dict = None,
        error_on_empty: bool = True,
    ) -> List[dict]:
        """
        Query for task run data necessary to initialize `TaskRunView` instances
        with `TaskRunView.from_task_run_data`.

        Args:
            - where (required): The Hasura `where` clause to filter by
            - order_by (optional): An optional Hasura `order_by` clause to order results
                by.
            - error_on_empty (optional): If `True` and no tasks are found, a `ValueError`
                will be raised.

        Returns:
           A list of dicts containing task run data
        """
        client = Client()

        query_args = {"where": where}
        if order_by is not None:
            query_args["order_by"] = order_by

        query = {
            "query": {
                with_args("task_run", query_args): {
                    "id": True,
                    "name": True,
                    "task": {"id": True, "slug": True},
                    "map_index": True,
                    "serialized_state": True,
                    "flow_run_id": True,
                }
            }
        }

        result = client.graphql(query)
        task_runs = result.get("data", {}).get("task_run", None)

        if task_runs is None:
            raise ValueError(
                f"Received bad result while querying for task runs where {where}: "
                f"{result}"
            )

        if not task_runs and error_on_empty:
            raise ValueError(
                f"No task runs found while querying for task runs where {where}"
            )

        return task_runs

    def __repr__(self) -> str:
        result = "<not loaded>" if self._result is NotLoaded else repr(self._result)
        return (
            f"{type(self).__name__}"
            "("
            + ", ".join(
                [
                    f"task_run_id={self.task_run_id!r}",
                    f"task_id={self.task_id!r}",
                    f"task_slug={self.task_slug!r}",
                    f"state={self.state}",
                    f"result={result}",
                ]
            )
            + ")"
        )

    def __eq__(self, other: Any) -> Any:
        if not isinstance(other, TaskRunView):
            return NotImplemented

        return other.task_run_id == self.task_run_id and other.state == self.state

    def get_latest(self) -> "TaskRunView":
        """
        Get the a new copy of this object with the latest data from the API. Cached
        result objects will be passed to the new object.

        This will not mutate the current object.

        Returns:
            A new instance of `TaskRunView`.
        """
        latest_view = self.from_task_run_id(task_run_id=self.task_run_id)
        latest_view._result = self._result
        return latest_view



import json
import time

import requests
import pendulum

from prefect import Task
from prefect.utilities.tasks import defaults_from_attrs


class FivetranSyncTask(Task):
    """
    Task for running Fivetran connector sync jobs.

    This task assumes the user is a Fivetran user (https://fivetran.com) who has
    successfully setup a connector and has access to the API credentials for
    that user (https://fivetran.com/account/settings, "API Config").

    Args:
        - connector_id (str, optional): Default connector id to use for sync jobs, if none is
            specified to `run`.
        - **kwargs (Any, optional): additional kwargs to pass to the base Task constructor
    """

    api_user_agent = "prefect/1.0.1"

    def __init__(self, connector_id: str = None, **kwargs):
        self.connector_id = connector_id
        super().__init__(**kwargs)

    @defaults_from_attrs("connector_id")
    def run(
        self,
        api_key: str,
        api_secret: str,
        connector_id: str = None,
        poll_status_every_n_seconds: int = 15,
        manual=True,
    ) -> dict:
        """
        Task run method for Fivetran connector syncs.

        An invocation of `run` will attempt to start a sync job for the specified `connector_id`. `run`
        will poll Fivetran for connector status, and will only complete when the sync has completed or
        when it receives an error status code from an API call.

        Args:
            - api_key (str): `API key` per https://fivetran.com/account/settings; should be secret!
            - api_secret (str): `API secret` per https://fivetran.com/account/settings; should be secret!
            - connector_id (str, optional): if provided, will overwrite value provided at init.
            - poll_status_every_n_seconds (int, optional): this task polls the Fivetran API for status,
                if provided this value will override the default polling time of 15 seconds.
            - manual (bool, optional): if provided, will overwrite Prefect's changes
                to the Fivetran connector's schedule, keeping it on Fivetran auto scheduling

        Returns:
            - dict: connector_id (str) and succeeded_at (timestamp str)
        """

        def parse_timestamp(api_time: str):
            """Returns either the pendulum-parsed actual timestamp or
            a very out-of-date timestamp if not set
            """
            return (
                pendulum.parse(api_time)
                if api_time is not None
                else pendulum.from_timestamp(-1)
            )

        if not connector_id:
            raise ValueError("Value for parameter `connector_id` must be provided.")
        if not api_key:
            raise ValueError("Value for parameter `api_key` must be provided.")
        if not api_secret:
            raise ValueError("Value for parameter `api_secret` must be provided.")

        URL_CONNECTOR: str = "https://api.fivetran.com/v1/connectors/{}".format(
            connector_id
        )

        self.logger.info(
            "Attempting start of Fivetran connector {}, sleep time set to {} seconds.".format(
                connector_id, poll_status_every_n_seconds
            )
        )

        headers = {"User-Agent": self.api_user_agent}

        # Automatically call `raise_for_status` on every request
        session = requests.Session()
        session.hooks = {"response": lambda r, *args, **kwargs: r.raise_for_status()}
        # Make sure connector configuration has been completed successfully and is not broken.
        resp = session.get(URL_CONNECTOR, auth=(api_key, api_secret), headers=headers)
        connector_details = resp.json()["data"]
        URL_LOGS = "https://fivetran.com/dashboard/connectors/{}/{}/logs".format(
            connector_details["service"], connector_details["schema"]
        )
        URL_SETUP = "https://fivetran.com/dashboard/connectors/{}/{}/setup".format(
            connector_details["service"], connector_details["schema"]
        )
        setup_state = connector_details["status"]["setup_state"]
        if setup_state != "connected":
            EXC_SETUP: str = (
                'Fivetran connector "{}" not correctly configured, status: {}; '
                + "please complete setup at {}"
            )
            raise ValueError(EXC_SETUP.format(connector_id, setup_state, URL_SETUP))
        # We need to know the previous job's completion time to know if the job succeeded or failed
        succeeded_at = parse_timestamp(connector_details["succeeded_at"])
        failed_at = parse_timestamp(connector_details["failed_at"])
        previous_completed_at = succeeded_at if succeeded_at > failed_at else failed_at
        # URL for connector logs within the UI
        self.logger.info(
            "Connector type: {}, connector schema: {}".format(
                connector_details["service"], connector_details["schema"]
            )
        )
        self.logger.info("Connectors logs at {}".format(URL_LOGS))

        # Set connector to manual sync mode, required to force sync through the API
        if manual:
            resp = session.patch(
                URL_CONNECTOR,
                data=json.dumps({"schedule_type": "manual"}),
                headers={"Content-Type": "application/json;version=2"},
                auth=(api_key, api_secret),
            )
        # Start connector sync
        resp = session.post(
            URL_CONNECTOR + "/force", auth=(api_key, api_secret), headers=headers
        )

        loop: bool = True
        while loop:
            resp = session.get(
                URL_CONNECTOR, auth=(api_key, api_secret), headers=headers
            )
            current_details = resp.json()["data"]
            # Failsafe, in case we missed a state transition – it is possible with a long enough
            # `poll_status_every_n_seconds` we could completely miss the 'syncing' state
            succeeded_at = parse_timestamp(current_details["succeeded_at"])
            failed_at = parse_timestamp(current_details["failed_at"])
            current_completed_at = (
                succeeded_at if succeeded_at > failed_at else failed_at
            )
            # The only way to tell if a sync failed is to check if its latest failed_at value
            # is greater than then last known "sync completed at" value.
            if failed_at > previous_completed_at:
                raise ValueError(
                    'Fivetran sync for connector "{}" failed; please see logs at {}'.format(
                        connector_id, URL_LOGS
                    )
                )
            # Started sync will spend some time in the 'scheduled' state before
            # transitioning to 'syncing'.
            # Capture the transition from 'scheduled' to 'syncing' or 'rescheduled',
            # and then back to 'scheduled' on completion.
            sync_state = current_details["status"]["sync_state"]
            self.logger.info(
                'Connector "{}" current sync_state = {}'.format(
                    connector_id, sync_state
                )
            )
            if current_completed_at > previous_completed_at:
                loop = False
            else:
                time.sleep(poll_status_every_n_seconds)

        return {
            "succeeded_at": succeeded_at.to_iso8601_string(),
            "connector_id": connector_id,
        }



import datetime
import json
import random
import re
import time
import uuid
import warnings
from pathlib import Path
from typing import TYPE_CHECKING, Any, Dict, List, Mapping, NamedTuple, Optional, Union
from urllib.parse import urljoin, urlparse

# if simplejson is installed, `requests` defaults to using it instead of json
# this allows the client to gracefully handle either json or simplejson
try:
    from simplejson.errors import JSONDecodeError
except ImportError:
    from json.decoder import JSONDecodeError

import pendulum
import toml
from slugify import slugify

import prefect
from prefect.exceptions import (
    AuthorizationError,
    ClientError,
    ObjectNotFoundError,
    VersionLockMismatchSignal,
)
from prefect.run_configs import RunConfig
from prefect.utilities.graphql import (
    EnumValue,
    GraphQLResult,
    compress,
    format_graphql_request_error,
    parse_graphql,
    with_args,
)
from prefect.utilities.logging import create_diagnostic_logger

if TYPE_CHECKING:
    import requests

    from prefect.core import Flow

JSONLike = Union[bool, dict, list, str, int, float, None]

# type definitions for GraphQL results


class TaskRunInfoResult(NamedTuple):
    # TODO: Deprecate this result in favor of `prefect.backend.TaskRun`
    id: str
    task_id: str
    task_slug: str
    version: int
    state: "prefect.engine.state.State"


class ProjectInfo(NamedTuple):
    id: str
    name: str


class FlowRunInfoResult(NamedTuple):
    # TODO: Deprecate this result in favor of `prefect.backend.FlowRun`
    id: str
    name: str
    flow_id: str
    project: ProjectInfo
    parameters: Dict[str, Any]
    context: Dict[str, Any]
    version: int
    scheduled_start_time: datetime.datetime
    state: "prefect.engine.state.State"
    task_runs: List[TaskRunInfoResult]


class Client:
    """
    Client for communication with Prefect Cloud

    If the arguments aren't specified the client initialization first checks the prefect
    configuration and if the server is not set there it checks the current context.

    Args:
        - api_server (str, optional): the URL to send all GraphQL requests to; if not
            provided, will be pulled from the current backend's `api` config
            variable
        - api_key (str, optional): a Prefect Cloud API key. If not provided, loaded
            from `config.cloud.api_key` or from the on disk cache from the
            `prefect auth` CLI
        - tenant_id (str, optional): the Prefect tenant to use. If not provided, loaded
            from `config.cloud.tenant_id` or the on disk cache from the
            `prefect auth` CLI
    """

    logger = create_diagnostic_logger("client.diagnostics")

    def __init__(
        self,
        api_server: str = None,
        api_key: str = None,
        tenant_id: str = None,
    ):
        self._attached_headers = {}  # type: Dict[str, str]

        # Hard-code the auth filepath location
        self._auth_file = Path(prefect.context.config.home_dir).absolute() / "auth.toml"

        # Note the default is `cloud.api` which is `cloud.endpoint` or `server.endpoint`
        # depending on the value of the `backend` key
        # This must be set before `load_auth_from_disk()` can be called
        self.api_server = api_server or prefect.context.config.cloud.api

        # Load the API key
        cached_auth = self.load_auth_from_disk()
        self.api_key: Optional[str] = (
            api_key
            or prefect.context.config.cloud.get("api_key")
            or cached_auth.get("api_key")
        )

        # mypy struggles with this attribute type if not created here
        self._tenant_id: Optional[str] = None

        # Load the tenant id
        # This assignment is validated by the `Client.tenant_id` setter
        self.tenant_id: Optional[str] = (
            tenant_id  # type: ignore
            or prefect.context.config.cloud.get("tenant_id")
            or cached_auth.get("tenant_id")
        )

        # If not set at this point, when `Client.tenant_id` is accessed the default
        # tenant will be loaded and used for future requests.

    # API key authentication -----------------------------------------------------------

    def _get_auth_tenant(self) -> str:
        """
        Get the current tenant associated with the API key being used. If the client has
        a specific tenant id set, this will verify that the given tenant id is
        compatible with the API key because the tenant will be attached to the request.
        """
        if not prefect.config.backend == "cloud":

            raise ValueError(
                "Authentication is only supported for Prefect Cloud. "
                "Your backend is set to {prefect.config.backend!r}"
            )

        if not self.api_key:
            raise ValueError("You have not set an API key for authentication.")

        response = self.graphql({"query": {"auth_info": "tenant_id"}})
        tenant_id = response.get("data", {}).get("auth_info", {}).get("tenant_id", "")

        if tenant_id == "":
            raise ClientError(
                "Unexpected response from the API while querying for the default "
                f"tenant: {response}"
            )

        elif tenant_id is None:
            # If the backend returns a `None` value tenant id, it indicates that an API
            # token was passed in as an API key
            raise AuthorizationError(
                "An API token was used as an API key. There is no tenant associated "
                "with API tokens. Use an API key for authentication."
            )

        return tenant_id

    def _get_default_server_tenant(self) -> str:
        if prefect.config.backend == "server":
            response = self.graphql({"query": {"tenant": {"id"}}})
            tenants = response.get("data", {}).get("tenant", None)
            if tenants is None:
                raise ClientError(
                    f"Unexpected response from the API while querying for tenants: {response}"
                )

            if not tenants:  # The user has not created a tenant yet
                raise ClientError(
                    "Your Prefect Server instance has no tenants. "
                    "Create a tenant with `prefect server create-tenant`"
                )

            return tenants[0].id

        elif prefect.config.backend == "cloud":
            raise ValueError(
                "Default tenants are determined by authentication in Prefect Cloud. "
                "See `_get_auth_tenant` instead."
            )
        else:
            raise ValueError("Unknown backend {prefect.config.backend!r}")

    def load_auth_from_disk(self) -> Dict[str, str]:
        """
        Get the stashed `api_key` and `tenant_id` for the current `api_server` from the
        disk cache if it exists. If it does not, an empty dict is returned.

        WARNING: This will not mutate the `Client`, you must use the returned dict
                 to set `api_key` and `tenant_id`. This is
        """
        if not self._auth_file.exists():
            return {}

        return toml.loads(self._auth_file.read_text()).get(self._api_server_slug, {})

    def save_auth_to_disk(self) -> None:
        """
        Write the current auth information to the disk cache under a header for the
        current `api_server`
        """
        # Load the current contents of the entire file
        contents = (
            toml.loads(self._auth_file.read_text()) if self._auth_file.exists() else {}
        )

        # Update the data for this API server
        contents[self._api_server_slug] = {
            "api_key": self.api_key,
            "tenant_id": self._tenant_id,
        }

        # Update the file, including a comment blurb
        self._auth_file.parent.mkdir(parents=True, exist_ok=True)
        self._auth_file.write_text(
            "# This file is auto-generated and should not be manually edited\n"
            "# Update the Prefect config or use the CLI to login instead\n\n"
            + toml.dumps(contents)
        )

    @property
    def _api_server_slug(self) -> str:
        """
        A slugified version of the API server's network location, used for loading
        and saving settings for different API servers.

        This should only be relevant for the 'cloud' backend since the 'server' backend
        does not save authentication.

        This should remain stable or users will not be able to load credentials
        """
        # Parse the server to drop the protocol
        netloc = urlparse(self.api_server).netloc
        # Then return the slugified contents, falling back to the full server if it
        # could not be parsed into a net location
        return slugify(netloc or self.api_server, regex_pattern=r"[^-\.a-z0-9]+")

    @property
    def tenant_id(self) -> str:
        """
        Retrieve the current tenant id the client is interacting with.

        If it is has not been explicitly set, the default tenant id will be retrieved
        """
        if not self._tenant_id:
            if prefect.config.backend == "cloud":
                self._tenant_id = self._get_auth_tenant()
            elif prefect.config.backend == "server":
                self._tenant_id = self._get_default_server_tenant()
            else:
                raise ValueError(f"Unknown backend setting {prefect.config.backend!r}")

        if not self._tenant_id:
            raise ClientError(
                "A tenant could not be determined. Please use `prefect auth status` "
                "to get information about your authentication and file an issue."
            )

        return self._tenant_id

    @tenant_id.setter
    def tenant_id(self, tenant_id: Union[str, uuid.UUID, None]) -> None:
        if not tenant_id:
            self._tenant_id = None
            return

        if not isinstance(tenant_id, uuid.UUID):
            try:
                tenant_id = uuid.UUID(tenant_id)
            except ValueError as exc:
                raise ValueError(
                    f"The `tenant_id` must be a valid UUID. Got {tenant_id!r}."
                ) from exc

        self._tenant_id = str(tenant_id)

    # ----------------------------------------------------------------------------------

    def create_tenant(self, name: str, slug: str = None) -> str:
        """
        Creates a new tenant.

        Note this route only works when run against Prefect Server.

        Args:
            - name (str): the name of the tenant to create
            - slug (str, optional): the slug of the tenant to create; defaults to name

        Returns:
            - str: the ID of the newly created tenant, or the ID of the currently active tenant

        Raises:
            - ValueError: if run against Prefect Cloud
        """
        if prefect.config.backend != "server":
            raise ValueError(
                "To create a tenant with Prefect Cloud, please signup at "
                "https://cloud.prefect.io/"
            )

        if slug is None:
            slug = slugify(name)

        tenant_info = self.graphql(
            {
                "mutation($input: create_tenant_input!)": {
                    "create_tenant(input: $input)": {"id"}
                }
            },
            variables=dict(input=dict(name=name, slug=slug)),
        )
        return tenant_info.data.create_tenant.id

    # -------------------------------------------------------------------------
    # Utilities

    def get(
        self,
        path: str,
        server: str = None,
        headers: dict = None,
        params: Dict[str, JSONLike] = None,
        api_key: str = None,
        retry_on_api_error: bool = True,
    ) -> dict:
        """
        Convenience function for calling the Prefect API with auth and GET request

        Args:
            - path (str): the path of the API url. For example, to GET
                http://prefect-server/v1/auth/login, path would be 'auth/login'.
            - server (str, optional): the server to send the GET request to;
                defaults to `self.api_server`
            - headers (dict, optional): Headers to pass with the request
            - params (dict): GET parameters
            - api_key (str): An api key for auth. Defaults to `client.api_key`.
            - retry_on_api_error (bool): whether the operation should be retried if the API returns
                an API_ERROR code

        Returns:
            - dict: Dictionary representation of the request made
        """
        response = self._request(
            method="GET",
            path=path,
            params=params,
            server=server,
            headers=headers,
            api_key=api_key,
            retry_on_api_error=retry_on_api_error,
        )
        if response.text:
            return response.json()
        else:
            return {}

    def post(
        self,
        path: str,
        server: str = None,
        headers: dict = None,
        params: Dict[str, JSONLike] = None,
        api_key: str = None,
        retry_on_api_error: bool = True,
    ) -> dict:
        """
        Convenience function for calling the Prefect API with auth and POST request

        Args:
            - path (str): the path of the API url. For example, to POST
                http://prefect-server/v1/auth/login, path would be 'auth/login'.
            - server (str, optional): the server to send the POST request to;
                defaults to `self.api_server`
            - headers(dict): headers to pass with the request
            - params (dict): POST parameters
            - api_key (str): An api key for auth. Defaults to `client.api_key`.
            - retry_on_api_error (bool): whether the operation should be retried if the API returns
                an API_ERROR code

        Returns:
            - dict: Dictionary representation of the request made
        """
        response = self._request(
            method="POST",
            path=path,
            params=params,
            server=server,
            headers=headers,
            api_key=api_key,
            retry_on_api_error=retry_on_api_error,
        )
        if response.text:
            return response.json()
        else:
            return {}

    def graphql(
        self,
        query: Any,
        raise_on_error: bool = True,
        headers: Dict[str, str] = None,
        variables: Mapping[str, JSONLike] = None,
        api_key: str = None,
        retry_on_api_error: bool = True,
    ) -> GraphQLResult:
        """
        Convenience function for running queries against the Prefect GraphQL API

        Args:
            - query (Any): A representation of a graphql query to be executed. It will be
                parsed by prefect.utilities.graphql.parse_graphql().
            - raise_on_error (bool): if True, a `ClientError` will be raised if the GraphQL
                returns any `errors`.
            - headers (dict): any additional headers that should be passed as part of the
                request
            - variables (dict): Variables to be filled into a query with the key being
                equivalent to the variables that are accepted by the query
            - api_key (str): An api key for auth. Defaults to `client.api_key`.
            - retry_on_api_error (bool): whether the operation should be retried if the API returns
                an API_ERROR code

        Returns:
            - dict: Data returned from the GraphQL query

        Raises:
            - ClientError if there are errors raised by the GraphQL mutation
        """
        result = self.post(
            path="",
            server=self.api_server,
            headers=headers,
            params=dict(query=parse_graphql(query), variables=json.dumps(variables)),
            api_key=api_key,
            retry_on_api_error=retry_on_api_error,
        )

        # TODO: It looks like this code is never reached because errors are raised
        #       in self._send_request by default
        if raise_on_error and "errors" in result:
            if "UNAUTHENTICATED" in str(result["errors"]):
                raise AuthorizationError(result["errors"])
            elif "Malformed Authorization header" in str(result["errors"]):
                raise AuthorizationError(result["errors"])
            elif (
                result["errors"][0].get("extensions", {}).get("code")
                == "VERSION_LOCKING_ERROR"
            ):
                raise VersionLockMismatchSignal(result["errors"])
            raise ClientError(result["errors"])
        else:
            return GraphQLResult(result)  # type: ignore

    def _send_request(
        self,
        session: "requests.Session",
        method: str,
        url: str,
        params: Dict[str, JSONLike] = None,
        headers: dict = None,
        rate_limit_counter: int = 1,
    ) -> "requests.models.Response":
        import requests

        if prefect.context.config.cloud.get("diagnostics") is True:
            self.logger.debug(f"Preparing request to {url}")
            clean_headers = {
                head: re.sub("Bearer .*", "Bearer XXXX", val)
                for head, val in headers.items()  # type: ignore
            }
            self.logger.debug(f"Headers: {clean_headers}")
            self.logger.debug(f"Request: {params}")
            start_time = time.time()

        if method == "GET":
            response = session.get(
                url,
                headers=headers,
                params=params,
                timeout=prefect.context.config.cloud.request_timeout,
            )
        elif method == "POST":
            response = session.post(
                url,
                headers=headers,
                json=params,
                timeout=prefect.context.config.cloud.request_timeout,
            )
        elif method == "DELETE":
            response = session.delete(
                url,
                headers=headers,
                timeout=prefect.context.config.cloud.request_timeout,
            )
        else:
            raise ValueError("Invalid method: {}".format(method))

        if prefect.context.config.cloud.get("diagnostics") is True:
            end_time = time.time()
            self.logger.debug(f"Response: {response.json()}")
            self.logger.debug(
                f"Request duration: {round(end_time - start_time, 4)} seconds"
            )

        # custom logic when encountering an API rate limit:
        # each time we encounter a rate limit, by default we sleep for
        # 3 minutes + random amount, where the random amount
        # is uniformly sampled from (0, 10 * 2 ** rate_limit_counter)
        # up to (0, 640), at which point an error is raised if the limit
        # is still being hit.
        #
        # sleep/backoff times can be customized in prefect.config.cloud.rate_limiting
        rate_limited = response.status_code == 429
        if (
            rate_limited
            and rate_limit_counter <= prefect.config.cloud.rate_limiting.max_retries
        ):
            jitter = (
                random.random()
                * prefect.config.cloud.rate_limiting.backoff_multiplier
                * (
                    prefect.config.cloud.rate_limiting.backoff_exponent
                    ** rate_limit_counter
                )
            )
            naptime = prefect.config.cloud.rate_limiting.backoff_s + jitter
            self.logger.warning(
                f"Rate limit encountered (attempt {rate_limit_counter}); sleeping for {naptime}s..."
            )
            time.sleep(naptime)
            response = self._send_request(
                session=session,
                method=method,
                url=url,
                params=params,
                headers=headers,
                rate_limit_counter=rate_limit_counter + 1,
            )

        # Check if request returned a successful status
        try:
            response.raise_for_status()
        except requests.HTTPError as exc:
            if response.status_code == 400 and params and "query" in params:
                # Create a custom-formatted err message for graphql errors which always
                # return a 400 status code and have "query" in the parameter dict
                try:
                    graphql_msg = format_graphql_request_error(response)
                except Exception:
                    # Fallback to a general message
                    graphql_msg = (
                        "This is likely caused by a poorly formatted GraphQL query or "
                        "mutation but the response could not be parsed for more details"
                    )
                raise ClientError(f"{exc}\n{graphql_msg}") from exc

            # Server-side and non-graphql errors will be raised without modification
            raise

        return response

    def _request(
        self,
        method: str,
        path: str,
        params: Dict[str, JSONLike] = None,
        server: str = None,
        headers: dict = None,
        api_key: str = None,
        retry_on_api_error: bool = True,
    ) -> "requests.models.Response":
        """
        Runs any specified request (GET, POST, DELETE) against the server

        Args:
            - method (str): The type of request to be made (GET, POST, DELETE)
            - path (str): Path of the API URL
            - params (dict, optional): Parameters used for the request
            - server (str, optional): The server to make requests against, base API
                server is used if not specified
            - headers (dict, optional): Headers to pass with the request
            - api_key (str): An api key for auth. Defaults to `client.api_key`.
            - retry_on_api_error (bool): whether the operation should be retried if the API returns
                an API_ERROR code

        Returns:
            - requests.models.Response: The response returned from the request

        Raises:
            - ClientError: on bad responses from the API
            - ValueError: if a method is specified outside of the accepted GET, POST, DELETE
            - requests.HTTPError: if a status code is returned that is not `200` or `401`
        """
        if server is None:
            server = self.api_server
        assert isinstance(server, str)  # mypy assert

        api_key = api_key or self.api_key

        # 'import requests' is expensive time-wise, we should do this just-in-time to keep
        # the 'import prefect' time low
        import requests

        url = urljoin(server, path.lstrip("/")).rstrip("/")

        params = params or {}

        headers = headers or {}
        if api_key:
            headers["Authorization"] = "Bearer {}".format(api_key)

        if self.api_key and self._tenant_id:
            # Attach a tenant id to the headers if using an API key since it can be
            # used accross tenants. API tokens cannot and do not need this header.
            headers["X-PREFECT-TENANT-ID"] = self._tenant_id

        headers["X-PREFECT-CORE-VERSION"] = str(prefect.__version__)

        if self._attached_headers:
            headers.update(self._attached_headers)

        session = requests.Session()
        retry_total = 6
        retries = requests.packages.urllib3.util.retry.Retry(
            total=retry_total,
            backoff_factor=1,
            status_forcelist=[500, 502, 503, 504],
            # typeshed is out of date with urllib3 and missing `allowed_methods`
            allowed_methods=["DELETE", "GET", "POST"],  # type: ignore
        )
        session.mount("https://", requests.adapters.HTTPAdapter(max_retries=retries))
        session.mount("http://", requests.adapters.HTTPAdapter(max_retries=retries))
        response = self._send_request(
            session=session, method=method, url=url, params=params, headers=headers
        )
        # parse the response
        try:
            json_resp = response.json()
        except JSONDecodeError as exc:
            if prefect.config.backend == "cloud" and "Authorization" not in headers:
                raise AuthorizationError(
                    "Malformed response received from Cloud - please ensure that you "
                    "are authenticated. See `prefect auth login --help`."
                ) from exc
            else:
                raise ClientError("Malformed response received from API.") from exc

        # check if there was an API_ERROR code in the response
        if "API_ERROR" in str(json_resp.get("errors")) and retry_on_api_error:
            success, retry_count = False, 0
            # retry up to six times
            while success is False and retry_count < 6:
                response = self._send_request(
                    session=session,
                    method=method,
                    url=url,
                    params=params,
                    headers=headers,
                )
                if "API_ERROR" in str(response.json().get("errors")):
                    retry_count += 1
                    time.sleep(0.25 * (2 ** (retry_count - 1)))
                else:
                    success = True

        return response

    def attach_headers(self, headers: dict) -> None:
        """
        Set headers to be attached to this Client

        Args:
            - headers (dict): A dictionary of headers to attach to this client. These headers
                get added on to the existing dictionary of headers.
        """
        self._attached_headers.update(headers)

    def switch_tenant(self, tenant_slug: str = None, tenant_id: str = None) -> str:
        """
        Switch this client to the given tenant by slug or tenant id.

        The client tenant will be updated but will not be saved to disk without an
        explicit call.

        Args:
            - tenant_slug (str): the tenant's slug
            - tenant_id (str): the tenant's id

        Returns:
            - str: The id of the tenant

        Raises:
            - ValueError: if no matching tenants are found
            - ValueError: both slug and id are provided
            - AuthenticationError: if the key is not valid for the given tenant

        """

        # Validate the given tenant id -------------------------------------------------

        if tenant_slug and tenant_id:
            raise ValueError(
                "Received both `tenant_slug` and `tenant_id`; only one is allowed."
            )

        if tenant_slug:
            tenant = self.graphql(
                {
                    "query($slug: String)": {
                        "tenant(where: {slug: { _eq: $slug } })": {"id"}
                    }
                },
                variables=dict(slug=tenant_slug),
            )
            if not tenant.data.tenant:
                raise ValueError(f"No matching tenant found for slug {tenant_slug!r}.")

            tenant_id = tenant.data.tenant[0].id

        if not tenant_id:
            raise ValueError("A `tenant_id` or `tenant_slug` must be provided.")

        self.tenant_id = tenant_id
        self._get_auth_tenant()

        return self.tenant_id

    def get_available_tenants(self) -> List[Dict]:
        """
        Returns a list of available tenants.

        Returns:
            - List[Dict]: a list of dictionaries containing the id, slug, and name of
                available tenants
        """
        result = self.graphql(
            {"query": {"tenant(order_by: {slug: asc})": {"id", "slug", "name"}}},
            api_key=self.api_key,
        )
        return result.data.tenant  # type: ignore

    # -------------------------------------------------------------------------
    # Actions
    # -------------------------------------------------------------------------

    def register(
        self,
        flow: "Flow",
        project_name: str = None,
        build: bool = True,
        set_schedule_active: bool = True,
        version_group_id: str = None,
        compressed: bool = True,
        no_url: bool = False,
        idempotency_key: str = None,
    ) -> str:
        """
        Register a new flow with Prefect Cloud.

        Args:
            - flow (Flow): a flow to register
            - project_name (str, optional): the project that should contain this flow.
            - build (bool, optional): if `True`, the flow's storage is built
                prior to serialization; defaults to `True`
            - set_schedule_active (bool, optional): if `False`, will set the schedule to
                inactive in the database to prevent auto-scheduling runs (if the Flow has a
                schedule).  Defaults to `True`. This can be changed later.
            - version_group_id (str, optional): the UUID version group ID to use for versioning
                this Flow in Cloud; if not provided, the version group ID associated with this
                Flow's project and name will be used.
            - compressed (bool, optional): if `True`, the serialized flow will be; defaults to
                `True` compressed
            - no_url (bool, optional): if `True`, the stdout from this function will not
                contain the URL link to the newly-registered flow in the Cloud UI
            - idempotency_key (optional, str): a key that, if matching the most recent
                registration call for this flow group, will prevent the creation of
                another flow version and return the existing flow id instead.

        Returns:
            - str: the ID of the newly-registered flow

        Raises:
            - ClientError: if the registration failed
        """
        required_parameters = {p for p in flow.parameters() if p.required}
        if flow.schedule is not None and required_parameters:
            required_names = {p.name for p in required_parameters}
            if not all(
                [
                    required_names <= set(c.parameter_defaults.keys())
                    for c in flow.schedule.clocks
                ]
            ):
                raise ClientError(
                    "Flows with required parameters can not be scheduled automatically."
                )
        if any(e.key for e in flow.edges) and flow.result is None:
            warnings.warn(
                "No result handler was specified on your Flow. Cloud features such as "
                "input caching and resuming task runs from failure may not work properly.",
                stacklevel=2,
            )
        if compressed:
            create_mutation = {
                "mutation($input: create_flow_from_compressed_string_input!)": {
                    "create_flow_from_compressed_string(input: $input)": {"id"}
                }
            }
        else:
            create_mutation = {
                "mutation($input: create_flow_input!)": {
                    "create_flow(input: $input)": {"id"}
                }
            }

        project = None

        if project_name is None:
            raise TypeError(
                "'project_name' is a required field when registering a flow."
            )

        query_project = {
            "query": {
                with_args("project", {"where": {"name": {"_eq": project_name}}}): {
                    "id": True
                }
            }
        }

        project = self.graphql(query_project).data.project  # type: ignore

        if not project:
            raise ValueError(
                "Project {} not found. Run `prefect create project '{}'` to create it.".format(
                    project_name, project_name
                )
            )

        serialized_flow = flow.serialize(build=build)  # type: Any

        # verify that the serialized flow can be deserialized
        try:
            prefect.serialization.flow.FlowSchema().load(serialized_flow)
        except Exception as exc:
            raise ValueError(
                "Flow could not be deserialized successfully. Error was: {}".format(
                    repr(exc)
                )
            ) from exc

        # prepare for batched registration
        serialized_tasks = serialized_flow.pop("tasks")
        serialized_edges = serialized_flow.pop("edges")

        if compressed:
            serialized_flow = compress(serialized_flow)

        inputs = dict(
            project_id=(project[0].id if project else None),
            serialized_flow=serialized_flow,
            # we don't want to begin scheduling work until all tasks are registered
            set_schedule_active=False,
            version_group_id=version_group_id,
        )
        # Add newly added inputs only when set for backwards compatibility
        if idempotency_key is not None:
            inputs.update(idempotency_key=idempotency_key)

        res = self.graphql(
            create_mutation,
            variables=dict(input=inputs),
            retry_on_api_error=False,
        )  # type: Any

        flow_id = (
            res.data.create_flow_from_compressed_string.id
            if compressed
            else res.data.create_flow.id
        )

        # batch register tasks and edges separately
        task_mutation = {
            "mutation($input: register_tasks_input!)": {
                "register_tasks(input: $input)": {"success"}
            }
        }
        edge_mutation = {
            "mutation($input: register_edges_input!)": {
                "register_edges(input: $input)": {"success"}
            }
        }

        # tasks in batches of 500
        start = 0
        batch_size = 500
        stop = start + batch_size

        while start <= len(serialized_tasks):
            task_batch = serialized_tasks[start:stop]
            inputs = dict(
                flow_id=flow_id,
                serialized_tasks=task_batch,
            )
            self.graphql(
                task_mutation,
                variables=dict(input=inputs),
            )
            start = stop
            stop += batch_size

        # edges in batches of 500
        start = 0
        batch_size = 500
        stop = start + batch_size

        while start <= len(serialized_edges):
            edge_batch = serialized_edges[start:stop]
            inputs = dict(
                flow_id=flow_id,
                serialized_edges=edge_batch,
            )
            self.graphql(
                edge_mutation,
                variables=dict(input=inputs),
            )
            start = stop
            stop += batch_size

        # finally, if requested, we turn on the schedule
        if set_schedule_active:
            schedule_mutation = {
                "mutation($input: set_schedule_active_input!)": {
                    "set_schedule_active(input: $input)": {"success"}
                }
            }
            inputs = dict(flow_id=flow_id)
            self.graphql(
                schedule_mutation,
                variables=dict(input=inputs),
            )

        if not no_url:
            # Query for flow group id
            res = self.graphql(
                {
                    "query": {
                        with_args("flow_by_pk", {"id": flow_id}): {"flow_group_id": ...}
                    }
                }
            )
            flow_group_id = res.get("data").flow_by_pk.flow_group_id

            # Generate direct link to Cloud flow
            flow_url = self.get_cloud_url("flow", flow_group_id)

            prefix = "└── "

            print("Flow URL: {}".format(flow_url))

            # Extra information to improve visibility
            if flow.run_config is not None:
                labels = sorted(flow.run_config.labels)
            else:
                labels = []
            msg = (
                f" {prefix}ID: {flow_id}\n"
                f" {prefix}Project: {project_name}\n"
                f" {prefix}Labels: {labels}"
            )
            print(msg)

        return flow_id

    def get_cloud_url(self, subdirectory: str, id: str) -> str:
        """
        Convenience method for creating Prefect Cloud URLs for a given subdirectory.

        Args:
            - subdirectory (str): the subdirectory to use (e.g., `"flow-run"`)
            - id (str): the ID of the page

        Returns:
            - str: the URL corresponding to the appropriate base URL, tenant slug, subdirectory
                and ID

        Example:

        ```python
        from prefect import Client

        client = Client()
        client.get_cloud_url("flow-run", "424242-ca-94611-111-55")
        # returns "https://cloud.prefect.io/my-tenant-slug/flow-run/424242-ca-94611-111-55"
        ```
        """

        # Search for matching cloud API because we can't guarantee that the backend config is set
        using_cloud_api = ".prefect.io" in prefect.config.cloud.api
        tenant_slug = self.get_default_tenant_slug()

        # For various API versions parse out `api-` for direct UI link
        base_url = (
            (
                re.sub("api-", "", prefect.config.cloud.api)
                if re.search("api-", prefect.config.cloud.api)
                else re.sub("api", "cloud", prefect.config.cloud.api)
            )
            if using_cloud_api
            else prefect.config.server.ui.endpoint
        )

        return "/".join([base_url.rstrip("/"), tenant_slug, subdirectory, id])

    def get_default_tenant_slug(self) -> str:
        """
        Get the default tenant slug for the currently authenticated user

        Returns:
            - str: the slug of the current default tenant for this user
        """
        res = self.graphql({"query": {"tenant": {"id", "slug"}}})

        tenants = res["data"]["tenant"]
        for tenant in tenants:
            # Return the slug if it matches the current tenant id
            if tenant.id == self.tenant_id:
                return tenant.slug

        raise ValueError(
            f"Failed to find current tenant {self.tenant_id!r} in result {res}"
        )

    def create_project(self, project_name: str, project_description: str = None) -> str:
        """
        Create a new project if a project with the name provided does not already exist

        Args:
            - project_name (str): the project that should be created
            - project_description (str, optional): the project description

        Returns:
            - str: the ID of the newly-created or pre-existing project

        Raises:
            - ClientError: if the project creation failed
        """
        project_mutation = {
            "mutation($input: create_project_input!)": {
                "create_project(input: $input)": {"id"}
            }
        }

        try:
            res = self.graphql(
                project_mutation,
                variables=dict(
                    input=dict(
                        name=project_name,
                        description=project_description,
                        tenant_id=self.tenant_id,
                    )
                ),
            )  # type: Any
        except ClientError as exc:
            if "'Uniqueness violation.'" in str(exc):
                project_query = {
                    "query": {
                        with_args(
                            "project", {"where": {"name": {"_eq": project_name}}}
                        ): {"id": True}
                    }
                }
                res = self.graphql(project_query)
                return res.data.project[0].id
            raise

        return res.data.create_project.id

    def delete_project(self, project_name: str) -> bool:
        """
        Delete a project

        Args:
            - project_name (str): the project that should be created

        Returns:
            - bool: True if project is deleted else False
        Raises:
            - ValueError: if the project is None or doesn't exist
        """

        if project_name is None:
            raise TypeError("'project_name' is a required field for deleting a project")

        query_project = {
            "query": {
                with_args("project", {"where": {"name": {"_eq": project_name}}}): {
                    "id": True
                }
            }
        }

        project = self.graphql(query_project).data.project

        if not project:
            raise ValueError("Project {} not found.".format(project_name))

        project_mutation = {
            "mutation($input: delete_project_input!)": {
                "delete_project(input: $input)": {"success"}
            }
        }

        delete_project = self.graphql(
            project_mutation, variables=dict(input=dict(project_id=project[0].id))
        )

        return delete_project.data.delete_project.success

    def create_flow_run(
        self,
        flow_id: str = None,
        context: dict = None,
        parameters: dict = None,
        run_config: RunConfig = None,
        labels: List[str] = None,
        scheduled_start_time: datetime.datetime = None,
        idempotency_key: str = None,
        run_name: str = None,
        version_group_id: str = None,
    ) -> str:
        """
        Create a new flow run for the given flow id.  If `start_time` is not provided, the flow
        run will be scheduled to start immediately.  If both `flow_id` and `version_group_id`
        are provided, only the `flow_id` will be used.

        Args:
            - flow_id (str, optional): the id of the Flow you wish to schedule
            - context (dict, optional): the run context
            - parameters (dict, optional): a dictionary of parameter values to pass to the flow run
            - run_config (RunConfig, optional): a run-config to use for this
                flow run, overriding any existing flow settings.
            - labels (List[str], optional): a list of labels to apply to the flow run
            - scheduled_start_time (datetime, optional): the time to schedule the execution
                for; if not provided, defaults to now
            - idempotency_key (str, optional): an idempotency key; if provided, this run will
                be cached for 24 hours. Any subsequent attempts to create a run with the same
                idempotency key will return the ID of the originally created run (no new run
                will be created after the first).  An error will be raised if parameters or
                context are provided and don't match the original.  Each subsequent request
                will reset the TTL for 24 hours.
            - run_name (str, optional): The name assigned to this flow run
            - version_group_id (str, optional): if provided, the unique unarchived flow within
                this version group will be scheduled to run.  This input can be used as a
                stable API for running flows which are regularly updated.

        Returns:
            - str: the ID of the newly-created flow run

        Raises:
            - ClientError: if the GraphQL query is bad for any reason
        """
        create_mutation = {
            "mutation($input: create_flow_run_input!)": {
                "create_flow_run(input: $input)": {"id": True}
            }
        }
        if not flow_id and not version_group_id:
            raise ValueError("One of flow_id or version_group_id must be provided")

        inputs = {}  # type: Dict[str, Any]
        if flow_id:
            inputs["flow_id"] = flow_id
        else:
            inputs["version_group_id"] = version_group_id

        # Always send an idempotency key, defaulting to a unique identifier
        # This will prevent multiple flow runs from being created if the client
        # retries this request
        inputs["idempotency_key"] = idempotency_key or uuid.uuid4().hex

        if parameters is not None:
            inputs["parameters"] = parameters
        if run_config is not None:
            inputs["run_config"] = run_config.serialize()
        if labels is not None:
            inputs["labels"] = labels
        if context is not None:
            inputs["context"] = context
        if scheduled_start_time is not None:
            inputs["scheduled_start_time"] = scheduled_start_time.isoformat()
        if run_name is not None:
            inputs["flow_run_name"] = run_name
        res = self.graphql(create_mutation, variables=dict(input=inputs))
        return res.data.create_flow_run.id  # type: ignore

    def get_flow_run_info(self, flow_run_id: str) -> FlowRunInfoResult:
        """
        Retrieves version and current state information for the given flow run.

        Args:
            - flow_run_id (str): the id of the flow run to get information for

        Returns:
            - GraphQLResult: an object representing information about the flow run

        Raises:
            - ClientError: if the GraphQL mutation is bad for any reason
        """
        from prefect.engine.state import State

        query = {
            "query": {
                with_args("flow_run_by_pk", {"id": flow_run_id}): {
                    "id": True,
                    "name": True,
                    "flow_id": True,
                    "parameters": True,
                    "context": True,
                    "version": True,
                    "scheduled_start_time": True,
                    "serialized_state": True,
                    # load all task runs except dynamic task runs
                    with_args("task_runs", {"where": {"map_index": {"_eq": -1}}}): {
                        "id": True,
                        "task": {"id": True, "slug": True},
                        "version": True,
                        "serialized_state": True,
                    },
                    "flow": {"project": {"name": True, "id": True}},
                }
            }
        }
        result = self.graphql(query).data.flow_run_by_pk  # type: ignore
        if result is None:
            raise ClientError('Flow run ID not found: "{}"'.format(flow_run_id))

        task_runs = [
            TaskRunInfoResult(
                id=tr.id,
                version=tr.version,
                task_id=tr.task.id,
                task_slug=tr.task.slug,
                state=State.deserialize(tr.serialized_state)
                if tr.serialized_state
                else prefect.engine.state.Pending(),
            )
            for tr in result.task_runs
        ]

        state = (
            prefect.engine.state.State.deserialize(result.serialized_state)
            if result.serialized_state
            else prefect.engine.state.Pending()
        )

        return FlowRunInfoResult(
            id=result.id,
            name=result.name,
            flow_id=result.flow_id,
            version=result.version,
            task_runs=task_runs,
            state=state,
            scheduled_start_time=pendulum.parse(result.scheduled_start_time),  # type: ignore
            project=ProjectInfo(
                id=result.flow.project.id, name=result.flow.project.name
            ),
            parameters=(
                {} if result.parameters is None else result.parameters.to_dict()
            ),
            context=({} if result.context is None else result.context.to_dict()),
        )

    def update_flow_run_heartbeat(self, flow_run_id: str) -> None:
        """
        Convenience method for heartbeating a flow run.

        Does NOT raise an error if the update fails.

        Args:
            - flow_run_id (str): the flow run ID to heartbeat

        """
        mutation = {
            "mutation": {
                with_args(
                    "update_flow_run_heartbeat", {"input": {"flow_run_id": flow_run_id}}
                ): {"success"}
            }
        }
        self.graphql(
            mutation,
            raise_on_error=True,
            headers={"X-PREFECT-HEARTBEAT-ID": flow_run_id},
        )

    def set_flow_run_name(self, flow_run_id: str, name: str) -> bool:
        """
        Set the name of a flow run.

        Args:
            - flow_run_id (str): the id of a flow run
            - name (str): a name for this flow run

        Returns:
            - bool: whether or not the flow run name was updated
        """
        mutation = {
            "mutation($input: set_flow_run_name_input!)": {
                "set_flow_run_name(input: $input)": {
                    "success": True,
                }
            }
        }

        result = self.graphql(
            mutation, variables=dict(input=dict(flow_run_id=flow_run_id, name=name))
        )

        return result.data.set_flow_run_name.success

    def get_flow_run_state(self, flow_run_id: str) -> "prefect.engine.state.State":
        """
        Retrieves the current state for a flow run.

        Args:
            - flow_run_id (str): the id for this flow run

        Returns:
            - State: a Prefect State object
        """
        query = {
            "query": {
                with_args("flow_run_by_pk", {"id": flow_run_id}): {
                    "serialized_state": True,
                }
            }
        }

        flow_run = self.graphql(query).data.flow_run_by_pk

        if not flow_run:
            raise ObjectNotFoundError(f"Flow run {flow_run_id!r} not found.")

        return prefect.engine.state.State.deserialize(flow_run.serialized_state)

    def set_flow_run_state(
        self,
        flow_run_id: str,
        state: "prefect.engine.state.State",
        version: int = None,
    ) -> "prefect.engine.state.State":
        """
        Sets new state for a flow run in the database.

        Args:
            - flow_run_id (str): the id of the flow run to set state for
            - state (State): the new state for this flow run
            - version (int, optional): the current version of the flow run state. This is optional
                but it can be supplied to enforce version-locking.

        Returns:
            - State: the state the current flow run should be considered in

        Raises:
            - ClientError: if the GraphQL mutation is bad for any reason
        """
        mutation = {
            "mutation($input: set_flow_run_states_input!)": {
                "set_flow_run_states(input: $input)": {
                    "states": {"id", "status", "message"}
                }
            }
        }

        serialized_state = state.serialize()

        result = self.graphql(
            mutation,
            variables=dict(
                input=dict(
                    states=[
                        dict(
                            state=serialized_state,
                            flow_run_id=flow_run_id,
                            version=version,
                        )
                    ]
                )
            ),
        )  # type: Any

        state_payload = result.data.set_flow_run_states.states[0]
        if state_payload.status == "QUEUED":
            # If appropriate, the state attribute of the Queued state can be
            # set by the caller of this method
            return prefect.engine.state.Queued(
                message=state_payload.get("message"),
                start_time=pendulum.now("UTC").add(
                    seconds=prefect.context.config.cloud.queue_interval
                ),
            )
        return state

    def get_latest_cached_states(
        self, task_id: str, cache_key: Optional[str], created_after: datetime.datetime
    ) -> List["prefect.engine.state.State"]:
        """
        Pulls all Cached states for the given task that were created after the provided date.

        Args:
            - task_id (str): the task id for this task run
            - cache_key (Optional[str]): the cache key for this Task's cache; if `None`, the
                task id alone will be used
            - created_after (datetime.datetime): the earliest date the state should have been
                created at

        Returns:
            - List[State]: a list of Cached states created after the given date
        """
        args = {
            "where": {
                "state": {"_eq": "Cached"},
                "state_timestamp": {"_gte": created_after.isoformat()},
            },
            "order_by": {"state_timestamp": EnumValue("desc")},
            "limit": 100,
        }  # type: Dict[str, Any]

        # if a cache key was provided, match it against all tasks
        if cache_key is not None:
            args["where"].update({"cache_key": {"_eq": cache_key}})
        # otherwise match against only this task, across all cache keys
        else:
            args["where"].update({"task_id": {"_eq": task_id}})

        query = {"query": {with_args("task_run", args): "serialized_state"}}
        result = self.graphql(query)  # type: Any
        deserializer = prefect.engine.state.State.deserialize
        valid_states = [
            deserializer(res.serialized_state) for res in result.data.task_run
        ]
        return valid_states

    def get_task_run_info(
        self, flow_run_id: str, task_id: str, map_index: Optional[int] = None
    ) -> TaskRunInfoResult:
        """
        Retrieves version and current state information for the given task run.

        Args:
            - flow_run_id (str): the id of the flow run that this task run lives in
            - task_id (str): the task id for this task run
            - map_index (int, optional): the mapping index for this task run; if
                `None`, it is assumed this task is _not_ mapped

        Returns:
            - NamedTuple: a tuple containing `id, task_id, version, state`

        Raises:
            - ClientError: if the GraphQL mutation is bad for any reason
        """

        mutation = {
            "mutation": {
                with_args(
                    "get_or_create_task_run_info",
                    {
                        "input": {
                            "flow_run_id": flow_run_id,
                            "task_id": task_id,
                            "map_index": -1 if map_index is None else map_index,
                        }
                    },
                ): {
                    "id": True,
                    "version": True,
                    "serialized_state": True,
                }
            }
        }
        result = self.graphql(mutation)  # type: Any

        if result is None:
            raise ClientError("Failed to create task run.")

        task_run_info = result.data.get_or_create_task_run_info

        state = (
            prefect.engine.state.State.deserialize(task_run_info.serialized_state)
            if task_run_info.serialized_state
            else prefect.engine.state.Pending()
        )

        return TaskRunInfoResult(
            id=task_run_info.id,
            task_id=task_id,
            task_slug="",
            version=task_run_info.version,
            state=state,
        )

    def set_task_run_name(self, task_run_id: str, name: str) -> bool:
        """
        Set the name of a task run

        Args:
            - task_run_id (str): the id of a task run
            - name (str): a name for this task run

        Returns:
            - bool: whether or not the task run name was updated
        """
        mutation = {
            "mutation($input: set_task_run_name_input!)": {
                "set_task_run_name(input: $input)": {
                    "success": True,
                }
            }
        }

        result = self.graphql(
            mutation, variables=dict(input=dict(task_run_id=task_run_id, name=name))
        )

        return result.data.set_task_run_name.success

    def cancel_flow_run(self, flow_run_id: str) -> bool:
        """
        Cancel the flow run by id

        Args:
            - flow_run_id (str): the id of the flow run

        Returns:
            - bool: whether or not the flow run was canceled
        """
        mutation = {
            "mutation($input: cancel_flow_run_input!)": {
                "cancel_flow_run(input: $input)": {
                    "state": True,
                }
            }
        }
        result = self.graphql(
            mutation, variables=dict(input=dict(flow_run_id=flow_run_id))
        )
        return result.data.cancel_flow_run.state

    def get_task_run_state(self, task_run_id: str) -> "prefect.engine.state.State":
        """
        Retrieves the current state for a task run.

        Args:
            - task_run_id (str): the id for this task run

        Returns:
            - State: a Prefect State object
        """
        query = {
            "query": {
                with_args("get_task_run_info", {"task_run_id": task_run_id}): {
                    "serialized_state": True,
                }
            }
        }

        task_run = self.graphql(query).data.get_task_run_info

        return prefect.engine.state.State.deserialize(task_run.serialized_state)

    def set_task_run_state(
        self,
        task_run_id: str,
        state: "prefect.engine.state.State",
        version: int = None,
        cache_for: datetime.timedelta = None,
    ) -> "prefect.engine.state.State":
        """
        Sets new state for a task run.

        Args:
            - task_run_id (str): the id of the task run to set state for
            - state (State): the new state for this task run
            - version (int, optional): the current version of the task run state. This is optional
                but it can be supplied to enforce version-locking.
            - cache_for (timedelta, optional): how long to store the result of this task for,
                using the serializer set in config; if not provided, no caching occurs

        Raises:
            - ClientError: if the GraphQL mutation is bad for any reason

        Returns:
            - State: the state the current task run should be considered in
        """
        mutation = {
            "mutation($input: set_task_run_states_input!)": {
                "set_task_run_states(input: $input)": {
                    "states": {"id", "status", "message"}
                }
            }
        }

        serialized_state = state.serialize()

        result = self.graphql(
            mutation,
            variables=dict(
                input=dict(
                    states=[
                        dict(
                            state=serialized_state,
                            task_run_id=task_run_id,
                            version=version,
                        )
                    ]
                )
            ),
        )  # type: Any
        state_payload = result.data.set_task_run_states.states[0]
        if state_payload.status == "QUEUED":
            # If appropriate, the state attribute of the Queued state can be
            # set by the caller of this method
            return prefect.engine.state.Queued(
                message=state_payload.get("message"),
                start_time=pendulum.now("UTC").add(
                    seconds=prefect.context.config.cloud.queue_interval
                ),
            )
        return state

    def set_secret(self, name: str, value: Any) -> None:
        """
        Set a secret with the given name and value.

        Args:
            - name (str): the name of the secret; used for retrieving the secret
                during task runs
            - value (Any): the value of the secret

        Raises:
            - ClientError: if the GraphQL mutation is bad for any reason
            - ValueError: if the secret-setting was unsuccessful
        """
        mutation = {
            "mutation($input: set_secret_input!)": {
                "set_secret(input: $input)": {"success"}
            }
        }

        result = self.graphql(
            mutation, variables=dict(input=dict(name=name, value=value))
        )  # type: Any

        if not result.data.set_secret.success:
            raise ValueError("Setting secret failed.")

    def get_task_tag_limit(self, tag: str) -> Optional[int]:
        """
        Retrieve the current task tag concurrency limit for a given tag.

        Args:
            - tag (str): the tag to update

        Raises:
            - ClientError: if the GraphQL query fails
        """
        query = {
            "query": {
                with_args("task_tag_limit", {"where": {"tag": {"_eq": tag}}}): {
                    "limit": True
                }
            }
        }

        result = self.graphql(query)  # type: Any
        if result.data.task_tag_limit:
            return result.data.task_tag_limit[0].limit
        else:
            return None

    def update_task_tag_limit(self, tag: str, limit: int) -> None:
        """
        Update the task tag concurrency limit for a given tag; requires tenant admin permissions.

        Args:
            - tag (str): the tag to update
            - limit (int): the concurrency limit to enforce on the tag; should be a value >= 0

        Raises:
            - ClientError: if the GraphQL mutation is bad for any reason
            - ValueError: if the tag limit-setting was unsuccessful, or if a bad limit was provided
        """
        if limit < 0:
            raise ValueError("Concurrency limits must be >= 0")

        mutation = {
            "mutation($input: update_task_tag_limit_input!)": {
                "update_task_tag_limit(input: $input)": {"id"}
            }
        }

        result = self.graphql(
            mutation, variables=dict(input=dict(tag=tag, limit=limit))
        )  # type: Any

        if not result.data.update_task_tag_limit.id:
            raise ValueError("Updating the task tag concurrency limit failed.")

    def delete_task_tag_limit(self, limit_id: str) -> None:
        """
        Deletes a given task tag concurrency limit; requires tenant admin permissions.

        Args:
            - limit_id (str): the ID of the tag to delete

        Raises:
            - ClientError: if the GraphQL mutation is bad for any reason
            - ValueError: if the tag deletion was unsuccessful, or if a bad tag ID was provided
        """
        mutation = {
            "mutation($input: delete_task_tag_limit_input!)": {
                "delete_task_tag_limit(input: $input)": {"success"}
            }
        }

        result = self.graphql(
            mutation, variables=dict(input=dict(limit_id=limit_id))
        )  # type: Any

        if not result.data.delete_task_tag_limit.success:
            raise ValueError("Deleting the task tag concurrency limit failed.")

    def write_run_logs(self, logs: List[Dict]) -> None:
        """
        Uploads a collection of logs to Cloud.

        Args:
            - logs (List[Dict]): a list of log entries to add

        Raises:
            - ValueError: if uploading the logs fail
        """
        mutation = {
            "mutation($input: write_run_logs_input!)": {
                "write_run_logs(input: $input)": {"success"}
            }
        }

        result = self.graphql(
            mutation, variables=dict(input=dict(logs=logs))
        )  # type: Any

        if not result.data.write_run_logs.success:
            raise ValueError("Writing logs failed.")

    def register_agent(
        self,
        agent_type: str,
        name: str = None,
        labels: List[str] = None,
        agent_config_id: str = None,
    ) -> str:
        """
        Register an agent with a backend API

        Args:
            - agent_type (str): The type of agent being registered
            - name: (str, optional): The name of the agent being registered
            - labels (List[str], optional): A list of any present labels on the agent
                being registered
            - agent_config_id (str, optional): The ID of an agent configuration to register with

        Returns:
            - The agent ID as a string
        """
        mutation = {
            "mutation($input: register_agent_input!)": {
                "register_agent(input: $input)": {"id"}
            }
        }

        result = self.graphql(
            mutation,
            variables=dict(
                input=dict(
                    type=agent_type,
                    name=name,
                    labels=labels or [],
                    tenant_id=self.tenant_id,
                    agent_config_id=agent_config_id,
                )
            ),
        )

        if not result.data.register_agent.id:
            raise ValueError("Error registering agent")

        return result.data.register_agent.id

    def get_agent_config(self, agent_config_id: str) -> dict:
        """
        Get agent config settings

        Args:
            - agent_config_id (str): The ID of an agent configuration to retrieve

        Returns:
            - dict: the agent configuration's `settings`
        """
        query = {
            "query": {
                with_args(
                    "agent_config", {"where": {"id": {"_eq": agent_config_id}}}
                ): {"settings": True}
            }
        }

        result = self.graphql(query)  # type: Any
        return result.data.agent_config[0].settings

    def create_task_run_artifact(
        self, task_run_id: str, kind: str, data: dict, tenant_id: str = None
    ) -> str:
        """
        Create an artifact that corresponds to a specific task run

        Args:
            - task_run_id (str): the task run id
            - kind (str): the artifact kind
            - data (dict): the artifact data
            - tenant_id (str, optional): the tenant id that this artifact belongs to. Defaults
                to the tenant ID linked to the task run

        Returns:
            - str: the task run artifact ID
        """
        mutation = {
            "mutation($input: create_task_run_artifact_input!)": {
                "create_task_run_artifact(input: $input)": {"id"}
            }
        }

        result = self.graphql(
            mutation,
            variables=dict(
                input=dict(
                    task_run_id=task_run_id, kind=kind, data=data, tenant_id=tenant_id
                )
            ),
        )

        artifact_id = result.data.create_task_run_artifact.id
        if not artifact_id:
            raise ValueError("Error creating task run artifact")

        return artifact_id

    def update_task_run_artifact(self, task_run_artifact_id: str, data: dict) -> None:
        """
        Update an artifact that corresponds to a specific task run

        Args:
            - task_run_artifact_id (str): the task run artifact id
            - data (dict): the artifact data
        """
        if task_run_artifact_id is None:
            raise ValueError(
                "The ID of an existing task run artifact must be provided."
            )

        mutation = {
            "mutation($input: update_task_run_artifact_input!)": {
                "update_task_run_artifact(input: $input)": {"success"}
            }
        }

        self.graphql(
            mutation,
            variables=dict(
                input=dict(task_run_artifact_id=task_run_artifact_id, data=data)
            ),
        )

    def delete_task_run_artifact(self, task_run_artifact_id: str) -> None:
        """
        Delete an artifact that corresponds to a specific task run

        Args:
            - task_run_artifact_id (str): the task run artifact id
        """
        if task_run_artifact_id is None:
            raise ValueError(
                "The ID of an existing task run artifact must be provided."
            )

        mutation = {
            "mutation($input: delete_task_run_artifact_input!)": {
                "delete_task_run_artifact(input: $input)": {"success"}
            }
        }

        self.graphql(
            mutation,
            variables=dict(input=dict(task_run_artifact_id=task_run_artifact_id)),
        )



from typing import Any, Dict, Iterable, List

import slugify

from prefect import config
from prefect.agent import Agent
from prefect.run_configs import VertexRun
from prefect.utilities.agent import get_flow_image, get_flow_run_command
from prefect.utilities.graphql import GraphQLResult


def get_client(client_options: dict) -> Any:
    from google.cloud import aiplatform

    return aiplatform.gapic.JobServiceClient(client_options=client_options)


class VertexAgent(Agent):
    """
    Agent which deploys flow runs as Vertex Training tasks.

    Args:
        - project (str): The project in which to submit the Vertex Jobs
            This does not necessarily need to be the same project as the where
            this agent is running, but the service account running the agent
            needs permissions to start Vertex Job in this project.
        - region_name (str, optional): Region the job is running in for the endpoint
        - service_account (str, optional): Service account to submit jobs as on Vertex
        - agent_config_id (str, optional): An optional agent configuration ID
            that can be used to set configuration based on an agent from a
            backend API. If set all configuration values will be pulled from
            the backend agent configuration.
        - name (str, optional): An optional name to give this agent. Can also
            be set through the environment variable `PREFECT__CLOUD__AGENT__NAME`.
            Defaults to "agent".
        - labels (List[str], optional): A list of labels, which are arbitrary
            string identifiers used by Prefect Agents when polling for work.
        - env_vars (dict, optional): A dictionary of environment variables and
            values that will be set on each flow run that this agent submits
            for execution.
        - max_polls (int, optional): Maximum number of times the agent will
            poll Prefect Cloud for flow runs; defaults to infinite.
        - agent_address (str, optional):  Address to serve internal api at.
            Currently this is just health checks for use by an orchestration
            layer. Leave blank for no api server (default).
        - no_cloud_logs (bool, optional): Disable logging to a Prefect backend
            for this agent and all deployed flow runs. Defaults to `False`.
    """

    def __init__(  # type: ignore
        self,
        project: str = None,
        region_name: str = None,
        service_account: str = None,
        agent_config_id: str = None,
        name: str = None,
        labels: Iterable[str] = None,
        env_vars: dict = None,
        max_polls: int = None,
        agent_address: str = None,
        no_cloud_logs: bool = None,
    ) -> None:
        super().__init__(
            agent_config_id=agent_config_id,
            name=name,
            labels=labels,
            env_vars=env_vars,
            max_polls=max_polls,
            agent_address=agent_address,
            no_cloud_logs=no_cloud_logs,
        )
        if not project:
            import google.auth

            _, project = google.auth.default()
            if project is None:
                # protect against any edge cases where project cannot be determined via auth default
                raise google.auth.exceptions.DefaultCredentialsError(
                    "Project could not be determined from environment"
                )

        self.project = project

        if not region_name:
            region_name = "us-central1"

        self.region_name = region_name
        self.service_account = service_account

        # Initialize client that will be used to create and send requests.
        # This client only needs to be created once, and can be reused for multiple requests.
        self.client_options = {
            "api_endpoint": f"{self.region_name}-aiplatform.googleapis.com"
        }
        self.vertex_client = get_client(self.client_options)

    def deploy_flow(self, flow_run: GraphQLResult) -> str:
        """
        Deploy a flow run as an Vertex task.

        Args:
            - flow_run (GraphQLResult): A GraphQLResult flow run object

        Returns:
            - str: Information about the deployment
        """
        taskdef = self.generate_task_definition(flow_run)
        parent = f"projects/{self.project}/locations/{self.region_name}"

        self.logger.debug(
            "Created task definition for flow %s",
            flow_run.flow.id,
        )

        resp = self.vertex_client.create_custom_job(parent=parent, custom_job=taskdef)
        self.logger.info(f"Created vertex job: {resp}")
        # The full name includes a projects key, to form the url we exclude it
        path = "/".join(resp.name.split("/")[2:]).replace("customJobs", "training")
        # return a url to navigate to the submitted flow
        url = f"https://console.cloud.google.com/vertex-ai/{path}"
        return url

    def generate_task_definition(self, flow_run: GraphQLResult) -> Dict[str, Any]:
        """Generate an Vertex task definition from a flow run

        Args:
            - flow_run (GraphQLResult): A flow run object

        Returns:
            - dict: a dictionary representation of an Vertex task definition
        """

        run_config = self._get_run_config(flow_run, VertexRun)
        assert isinstance(run_config, VertexRun)  # mypy

        image = get_flow_image(flow_run)
        job_name = slugify.slugify(
            flow_run.flow.name + "-" + flow_run.name,
            max_length=255,
            word_boundary=True,
            save_order=True,
        )
        machine_type = run_config.machine_type

        command = get_flow_run_command(flow_run)
        env = self.populate_env_vars(flow_run)
        env_list = self._to_env_list(env)

        # Start with a default taskdef
        taskdef = {
            "display_name": job_name,
            "job_spec": {
                "worker_pool_specs": [
                    {"machine_spec": {"machine_type": machine_type}, "replica_count": 1}
                ]
            },
        }  # type: Dict[str, Any]

        if run_config.worker_pool_specs is not None:
            taskdef["job_spec"]["worker_pool_specs"] = run_config.worker_pool_specs

        if run_config.network is not None:
            taskdef["job_spec"]["network"] = run_config.network

        if run_config.service_account is not None:
            taskdef["job_spec"]["service_account"] = run_config.service_account
        else:
            taskdef["job_spec"]["service_account"] = self.service_account

        if run_config.scheduling is not None:
            taskdef["job_spec"]["scheduling"] = run_config.scheduling

        # We always set the container spec on the zeroth pool spec to ensure it will run the flow
        taskdef["job_spec"]["worker_pool_specs"][0]["container_spec"] = {
            "image_uri": image,
            "command": command.split(),
            "args": [],
            "env": env_list,
        }

        return taskdef

    def populate_env_vars(self, flow_run: GraphQLResult) -> dict:
        """
        Populate metadata and variables in the environment variables for a flow run
        Args:
            - flow_run (GraphQLResult): A flow run object
            - image (str): The image for this flow
            - run_config (VertexRun, optional): The `run_config` for the flow, if any.
        Returns:
            - dict: a dictionary representing the populated environment variables
        """

        run_config = self._get_run_config(flow_run, VertexRun)

        env = {}
        # Populate environment variables, later sources overriding

        # 1. Logging level from config
        # Default to the config logging level, allowing it to be overriden
        # by later config soruces
        env.update({"PREFECT__LOGGING__LEVEL": config.logging.level})

        # 2. Values set on the agent via `--env`
        env.update(self.env_vars)

        # 3. Values set on a RunConfig (if present)
        if run_config is not None and run_config.env is not None:
            env.update(run_config.env)

        # 4. Allow logs to be disabled by flow run settings but not enabled
        if not self.log_to_cloud or "PREFECT__CLOUD__SEND_FLOW_RUN_LOGS" not in env:
            env["PREFECT__CLOUD__SEND_FLOW_RUN_LOGS"] = str(self.log_to_cloud).lower()

        # 5. Non-overrideable required env vars
        env.update(
            {
                "PREFECT__BACKEND": config.backend,
                "PREFECT__CLOUD__API": config.cloud.api,
                "PREFECT__CLOUD__API_KEY": self.flow_run_api_key or "",
                "PREFECT__CLOUD__TENANT_ID": (
                    # Providing a tenant id is only necessary when authenticating
                    self.client.tenant_id
                    if self.flow_run_api_key
                    else ""
                ),
                "PREFECT__CLOUD__AGENT__LABELS": str(self.labels),
                "PREFECT__CONTEXT__FLOW_RUN_ID": flow_run.id,  # type: ignore
                "PREFECT__CONTEXT__FLOW_ID": flow_run.flow.id,  # type: ignore
                "PREFECT__CLOUD__USE_LOCAL_SECRETS": "false",
                "PREFECT__ENGINE__FLOW_RUNNER__DEFAULT_CLASS": "prefect.engine.cloud.CloudFlowRunner",
                # Backwards compatibility variable for containers on Prefect <0.15.0
                "PREFECT__LOGGING__LOG_TO_CLOUD": str(self.log_to_cloud).lower(),
                # Backwards compatibility variable for containers on Prefect <1.0.0
                "PREFECT__CLOUD__AUTH_TOKEN": self.flow_run_api_key or "",
            }
        )
        return env

    def _to_env_list(self, env: dict) -> List[dict]:
        """Convert from dictionary to the expected vertex env list format"""
        env_list = []
        for k, v in env.items():
            env_list.append({"name": k, "value": v})
        return env_list



from prefect.executors.base import Executor
from prefect.engine.executors.dask import DaskExecutor, LocalDaskExecutor
from prefect.engine.executors.local import LocalExecutor

__all__ = ["DaskExecutor", "Executor", "LocalDaskExecutor", "LocalExecutor"]



"""
This module contains a collection of tasks for interacting with the [Toloka](https://toloka.ai/) API.

All Toloka-related tasks can be authenticated using the `TOLOKA_TOKEN` Prefect Secret that should contain a Toloka API OAuth token. See [Third Party Authentication](../../../orchestration/recipes/third_party_auth.html) for more information.
"""
try:
    from .helpers import *
    from .operations import *
except ImportError as err:
    raise ImportError(
        'Using `prefect.tasks.tasks` requires Prefect to be installed with the "toloka" extra.'
    ) from err



import filecmp
import json
import os
import copy
import re
import shutil
import sys
import tempfile
import textwrap
import uuid
import warnings
from pathlib import PurePosixPath, Path
from typing import TYPE_CHECKING, Any, Iterable, List, Union

import pendulum
from slugify import slugify

import prefect
from prefect.storage import Storage
from prefect.utilities.storage import (
    extract_flow_from_file,
    flow_from_bytes_pickle,
    flow_to_bytes_pickle,
)

if TYPE_CHECKING:
    import docker


def multiline_indent(string: str, spaces: int) -> str:
    """
    Utility to indent all but the first line in a string to the specified level,
    especially useful in textwrap.dedent calls where multiline strings are formatted in
    """
    return string.replace("\n", "\n" + " " * spaces)


class Docker(Storage):
    """
    Docker storage provides a mechanism for storing Prefect flows in Docker
    images and optionally pushing them to a registry.

    A user specifies a `registry_url`, `base_image` and other optional
    dependencies (e.g., `python_dependencies`) and `build()` will create a
    temporary Dockerfile that is used to build the image.

    Note that the `base_image` must be capable of `pip` installing.  Note that
    registry behavior with respect to image names can differ between providers -
    for example, Google's GCR registry allows for registry URLs of the form
    `gcr.io/my-registry/subdir/my-image-name` whereas DockerHub requires the
    registry URL to be separate from the image name.

    Custom modules can be packaged up during build by attaching the files and
    setting the `PYTHONPATH` to the location of those files. Otherwise the
    modules can be set independently when using a custom base image prior to the
    build here.

    ```python
    Docker(
        files={
            # absolute path source -> destination in image
            "/Users/me/code/mod1.py": "/modules/mod1.py",
            "/Users/me/code/mod2.py": "/modules/mod2.py",
        },
        env_vars={
            # append modules directory to PYTHONPATH
            "PYTHONPATH": "$PYTHONPATH:modules/"
        },
    )
    ```

    Args:
        - registry_url (str, optional): URL of a registry to push the image to;
            image will not be pushed if not provided
        - base_image (str, optional): the base image for this when building this
            image (e.g. `python:3.7`), defaults to the `prefecthq/prefect` image
            matching your python version and prefect core library version used
            at runtime.
        - dockerfile (str, optional): a path to a Dockerfile to use in building
            this storage; note that, if provided, your present working directory
            will be used as the build context
        - python_dependencies (List[str], optional): list of pip installable
            dependencies for the image
        - image_name (str, optional): name of the image to use when building,
            populated with a UUID after build
        - image_tag (str, optional): tag of the image to use when building,
            populated with a UUID after build
        - env_vars (dict, optional): a dictionary of environment variables to
            use when building
        - files (dict, optional): a dictionary of files or directories to copy into
            the image when building. Takes the format of `{'src': 'dest'}`
        - dockerignore (str, optional): an optional Path to a `dockerignore` file.
            when used with the `files` argument, the specified `dockerignore` will be
            included in the build context
        - prefect_version (str, optional): an optional branch, tag, or commit
            specifying the version of prefect you want installed into the container;
            defaults to the version you are currently using or `"master"` if your
            version is ahead of the latest tag
        - local_image (bool, optional): an optional flag whether or not to use a
            local docker image, if True then a pull will not be attempted
        - ignore_healthchecks (bool, optional): if True, the Docker healthchecks
            are not added to the Dockerfile. If False (default), healthchecks
            are included.
        - base_url (str, optional): a URL of a Docker daemon to use when for
            Docker related functionality.  Defaults to DOCKER_HOST env var if not set
        - tls_config (Union[bool, docker.tls.TLSConfig], optional): a TLS configuration to pass
            to the Docker client.
            [Documentation](https://docker-py.readthedocs.io/en/stable/tls.html#docker.tls.TLSConfig)
        - build_kwargs (dict, optional): Additional keyword arguments to pass to Docker's build
            step. [Documentation](https://docker-py.readthedocs.io/en/stable/
            api.html#docker.api.build.BuildApiMixin.build)
        - prefect_directory (str, optional): Path to the directory where prefect configuration/flows
             should be stored inside the Docker image. Defaults to `/opt/prefect`.
        - path (str, optional): a direct path to the location of the flow file in the Docker image
            if `stored_as_script=True`.
        - stored_as_script (bool, optional): boolean for specifying if the flow has been stored
            as a `.py` file. Defaults to `False`
        - extra_dockerfile_commands (list[str], optional): list of Docker build commands
            which are injected at the end of generated DockerFile (before the health checks).
            Defaults to `None`
        - **kwargs (Any, optional): any additional `Storage` initialization options

    Raises:
        - ValueError: if both `base_image` and `dockerfile` are provided

    """

    def __init__(
        self,
        registry_url: str = None,
        base_image: str = None,
        dockerfile: str = None,
        dockerignore: str = None,
        python_dependencies: List[str] = None,
        image_name: str = None,
        image_tag: str = None,
        env_vars: dict = None,
        files: dict = None,
        prefect_version: str = None,
        local_image: bool = False,
        ignore_healthchecks: bool = False,
        base_url: str = None,
        tls_config: Union[bool, "docker.tls.TLSConfig"] = False,
        build_kwargs: dict = None,
        prefect_directory: str = "/opt/prefect",
        path: str = None,
        stored_as_script: bool = False,
        extra_dockerfile_commands: List[str] = None,
        **kwargs: Any,
    ) -> None:
        self.registry_url = registry_url
        if sys.platform == "win32":
            default_url = "npipe:////./pipe/docker_engine"
        else:
            default_url = "unix://var/run/docker.sock"
        self.image_name = image_name
        self.image_tag = image_tag
        self.python_dependencies = python_dependencies or []
        self.python_dependencies.append("wheel")

        self.prefect_directory = prefect_directory
        self.path = path

        self.env_vars = env_vars or {}
        self.env_vars.setdefault(
            "PREFECT__USER_CONFIG_PATH", "{}/config.toml".format(self.prefect_directory)
        )

        self.files = files or {}
        self.local_image = local_image
        self.installation_commands = []  # type: List[str]
        self.ignore_healthchecks = ignore_healthchecks

        self.base_url = base_url or os.environ.get("DOCKER_HOST", default_url)
        self.tls_config = tls_config
        self.build_kwargs = copy.copy(build_kwargs) or {}
        self.rm_build_kwarg = self.build_kwargs.pop("rm", True)
        self.extra_dockerfile_commands = extra_dockerfile_commands

        version = prefect.__version__.split("+")
        if prefect_version is None:
            self.prefect_version = (
                "master"
                # The release candidate is a special development version
                if len(version) > 1 and not version[0].endswith("rc0")
                else version[0]
            )
        else:
            self.prefect_version = prefect_version

        if base_image is None and dockerfile is None:
            python_version = "{}.{}".format(
                sys.version_info.major, sys.version_info.minor
            )
            if re.match(r"^[0-9]+\.[0-9]+\.[0-9]+$", self.prefect_version) is not None:
                self.base_image = "prefecthq/prefect:{}-python{}".format(
                    self.prefect_version, python_version
                )
            elif self.prefect_version.endswith("rc0"):
                # Development release candidate
                self.base_image = f"prefecthq/prefect:{self.prefect_version}"
            elif (
                re.match(r"^[0-9]+\.[0-9]+rc[0-9]+$", self.prefect_version) is not None
            ):
                # Actual release candidate
                self.base_image = "prefecthq/prefect:{}-python{}".format(
                    self.prefect_version, python_version
                )
            else:
                # create an image from python:*-slim directly
                self.base_image = "python:{}-slim".format(python_version)
                self.installation_commands.append(
                    "apt update && apt install -y gcc git make && rm -rf /var/lib/apt/lists/*"
                )
        elif base_image and dockerfile:
            raise ValueError(
                "Only one of `base_image` and `dockerfile` can be provided."
            )
        else:
            self.base_image = base_image  # type: ignore

        self.dockerfile = dockerfile
        self.dockerignore = dockerignore
        # we should always try to install prefect, unless it is already installed. We can't
        # determine this until image build time.
        self.installation_commands.append(
            f"pip show prefect || "
            f"pip install git+https://github.com/PrefectHQ/prefect.git"
            f"@{self.prefect_version}#egg=prefect[all_orchestration_extras]"
        )

        not_absolute = [
            file_path for file_path in self.files if not os.path.isabs(file_path)
        ]
        if not_absolute:
            raise ValueError(
                (
                    "Provided paths {} are not absolute file paths, please provide "
                    "absolute paths only."
                ).format(", ".join(not_absolute))
            )
        super().__init__(stored_as_script=stored_as_script, **kwargs)

    def add_flow(self, flow: "prefect.core.flow.Flow") -> str:
        """
        Method for adding a new flow to this Storage object.

        Args:
            - flow (Flow): a Prefect Flow to add

        Returns:
            - str: the location of the newly added flow in this Storage object
        """
        if flow.name in self:
            raise ValueError(
                'Name conflict: Flow with the name "{}" is already present in this storage.'.format(
                    flow.name
                )
            )
        flow_path = self.path or "{}/flows/{}.prefect".format(
            self.prefect_directory, slugify(flow.name)
        )
        self.flows[flow.name] = flow_path
        self._flows[flow.name] = flow  # needed prior to build
        return flow_path

    def get_flow(self, flow_name: str) -> "prefect.core.flow.Flow":
        """
        Given a flow name within this Storage object, load and return the Flow.

        Args:
            - flow_name (str): the name of the flow to return.

        Returns:
            - Flow: the requested flow
        """
        if flow_name not in self.flows:
            raise ValueError("Flow is not contained in this Storage")
        flow_location = self.flows[flow_name]

        if self.stored_as_script:
            return extract_flow_from_file(file_path=flow_location, flow_name=flow_name)

        with open(flow_location, "rb") as f:
            return flow_from_bytes_pickle(f.read())

    @property
    def name(self) -> str:
        """
        Full name of the Docker image.
        """
        if None in [self.image_name, self.image_tag]:
            raise ValueError(
                "Docker storage is missing required fields image_name and image_tag"
            )

        return "{}:{}".format(
            PurePosixPath(self.registry_url or "", self.image_name),  # type: ignore
            self.image_tag,  # type: ignore
        )

    def build(self, push: bool = True) -> "Storage":
        """
        Build the Docker storage object.  If image name and tag are not set,
        they will be autogenerated.

        Args:
            - push (bool, optional): Whether or not to push the built Docker image, this
                requires the `registry_url` to be set

        Returns:
            - Docker: a new Docker storage object that contains information about how and
                where the flow is stored. Image name and tag are generated during the
                build process.

        Raises:
            - InterruptedError: if either pushing or pulling the image fails
        """
        if len(self.flows) != 1:
            self.image_name = self.image_name or str(uuid.uuid4())
        else:
            self.image_name = self.image_name or slugify(list(self.flows.keys())[0])

        self.image_tag = self.image_tag or slugify(pendulum.now("utc").isoformat())
        self._build_image(push=push)
        return self

    def _build_image(self, push: bool = True) -> tuple:
        """
        Build a Docker image using the docker python library.

        Args:
            - push (bool, optional): Whether or not to push the built Docker image, this
                requires the `registry_url` to be set

        Returns:
            - tuple: generated UUID strings `image_name`, `image_tag`

        Raises:
            - ValueError: if the image fails to build
            - InterruptedError: if either pushing or pulling the image fails
        """
        assert isinstance(self.image_name, str), "Image name must be provided"
        assert isinstance(self.image_tag, str), "An image tag must be provided"

        # Make temporary directory to hold serialized flow, healthcheck script, and dockerfile
        # note that if the user provides a custom dockerfile, we create the temporary directory
        # within the current working directory to preserve their build context
        with tempfile.TemporaryDirectory(
            dir="." if self.dockerfile else None
        ) as tempdir:
            # Build the dockerfile
            if self.base_image and not self.local_image:
                self.pull_image()

            dockerfile_path = self.create_dockerfile_object(directory=tempdir)
            client = self._get_client()

            # Verify that a registry url has been provided for images that should be
            # pushed
            if self.registry_url:
                full_name = str(PurePosixPath(self.registry_url, self.image_name))
            elif push is True:
                warnings.warn(
                    "This Docker storage object has no `registry_url`, and "
                    "will not be pushed.",
                    UserWarning,
                    stacklevel=2,
                )
                full_name = self.image_name
            else:
                full_name = self.image_name

            # Use the docker client to build the image
            self.logger.info("Building the flow's Docker storage...")

            if sys.platform == "win32":
                # problem with docker and relative paths only on windows
                dockerfile_path = os.path.abspath(dockerfile_path)

            output = client.build(
                path="." if self.dockerfile else tempdir,
                dockerfile=dockerfile_path,
                tag="{}:{}".format(full_name, self.image_tag),
                rm=self.rm_build_kwarg,
                **self.build_kwargs,
            )
            self._parse_generator_output(output)

            if len(client.images(name="{}:{}".format(full_name, self.image_tag))) == 0:
                raise ValueError(
                    "Your docker image failed to build!  Your flow might have "
                    "failed one of its deployment health checks - please ensure "
                    "that all necessary files and dependencies have been included."
                )

            # Push the image if requested
            if push and self.registry_url:
                self.push_image(full_name, self.image_tag)

                # Remove the image locally after being pushed
                client.remove_image(
                    image="{}:{}".format(full_name, self.image_tag), force=True
                )

        return self.image_name, self.image_tag

    ########################
    # Dockerfile Creation
    ########################

    def create_dockerfile_object(self, directory: str) -> str:
        """
        Writes a dockerfile to the provided directory using the specified
        arguments on this Docker storage object.

        In order for the docker python library to build a container it needs a
        Dockerfile that it can use to define the container. This function takes the
        specified arguments then writes them to a temporary file called Dockerfile.

        *Note*: if `files` are added to this container, they will be copied to this
        directory as well.

        Args:
            - directory (str): A directory where the Dockerfile will be created

        Returns:
            - str: the absolute file path to the Dockerfile
        """
        # Either load the base commands from the specified dockerfile or define a FROM
        if self.dockerfile:
            with open(self.dockerfile, "r") as contents:
                base_commands = contents.read()
        else:
            base_commands = "FROM {base_image}".format(base_image=self.base_image)

        # Generate ENV variables to load into the image
        env_vars = ""
        if self.env_vars:
            # Format with repr to get proper quoting
            formatted_vars = [f"{k}={v!r}" for k, v in self.env_vars.items()]
            env_vars = "ENV " + " \\\n    ".join(formatted_vars)

        # Generate single pip install command for python dependencies
        pip_installs = "RUN pip install "
        if self.python_dependencies:
            for dependency in self.python_dependencies:
                pip_installs += "'{}' ".format(dependency)

        # Write all install-time commands that should be run in the image
        installation_commands = ""
        for cmd in self.installation_commands:
            installation_commands += "RUN {}\n".format(cmd)

        # Copy user specified files into the image
        copy_files = ""
        if self.files:
            # copy dockerignore if provided
            if self.dockerignore:
                shutil.copyfile(
                    src=Path(self.dockerignore).expanduser().resolve(),
                    dst=Path(directory) / ".dockerignore",
                )
            for src, dest in self.files.items():
                fname = os.path.basename(src)
                full_fname = os.path.join(directory, fname)
                if os.path.exists(full_fname) and filecmp.cmp(src, full_fname) is False:
                    raise ValueError(
                        "File {fname} already exists in {directory}".format(
                            fname=full_fname, directory=directory
                        )
                    )
                else:
                    if os.path.isdir(src):
                        shutil.copytree(
                            src=src, dst=full_fname, symlinks=False, ignore=None
                        )
                    else:
                        shutil.copy2(src=src, dst=full_fname)
                copy_files += "COPY {fname} {dest}\n".format(
                    fname=full_fname.replace("\\", "/") if self.dockerfile else fname,
                    dest=dest,
                )

        # Write all flows to file and load into the image
        copy_flows = ""
        if not self.stored_as_script:
            for flow_name, flow_location in self.flows.items():
                clean_name = slugify(flow_name)
                flow_path = os.path.join(directory, "{}.flow".format(clean_name))
                with open(flow_path, "wb") as f:
                    f.write(flow_to_bytes_pickle(self._flows[flow_name]))
                copy_flows += "COPY {source} {dest}\n".format(
                    source=(
                        flow_path.replace("\\", "/")
                        if self.dockerfile
                        else "{}.flow".format(clean_name)
                    ),
                    dest=flow_location,
                )
        else:
            if not self.path:
                raise ValueError(
                    "A `path` must be provided to show where flow `.py` file is stored in the image."
                )

        # Write final extra user commands that should be run in the image
        final_commands = (
            ""
            if self.extra_dockerfile_commands is None
            else "\n".join(self.extra_dockerfile_commands)
        )

        # Write a healthcheck script into the image
        with open(
            os.path.join(os.path.dirname(__file__), "_healthcheck.py"), "r"
        ) as healthscript:
            healthcheck = healthscript.read()

        healthcheck_loc = os.path.join(directory, "healthcheck.py")
        with open(healthcheck_loc, "w") as health_file:
            health_file.write(healthcheck)

        # Escape the healthcheck location
        healthcheck_loc = (
            healthcheck_loc.replace("\\", "/") if self.dockerfile else "healthcheck.py"
        )

        # Generate the command to run the healthcheck
        healthcheck_run = ""
        if not self.ignore_healthchecks:
            flow_file_paths = ", ".join(['"{}"'.format(k) for k in self.flows.values()])
            python_version = (sys.version_info.major, sys.version_info.minor)
            healthcheck_run = (
                f"RUN python {self.prefect_directory}/healthcheck.py "
                f"'[{flow_file_paths}]' '{python_version}'"
            )

        file_contents = textwrap.dedent(
            f"""
            {multiline_indent(base_commands, 12)}
            {multiline_indent(env_vars, 12)}

            RUN pip install pip --upgrade
            {multiline_indent(installation_commands, 12)}
            {pip_installs}

            RUN mkdir -p {self.prefect_directory}/
            {multiline_indent(copy_flows, 12)}
            COPY {healthcheck_loc} {self.prefect_directory}/healthcheck.py
            {multiline_indent(copy_files, 12)}

            {multiline_indent(final_commands, 12)}
            {healthcheck_run}
            """
        )

        dockerfile_path = os.path.join(directory, "Dockerfile")
        with open(dockerfile_path, "w+") as dockerfile:
            dockerfile.write(file_contents)
        return dockerfile_path

    ########################
    # Docker Utilities
    ########################

    def _get_client(self) -> "docker.APIClient":
        # 'import docker' is expensive time-wise, we should do this just-in-time to keep
        # the 'import prefect' time low
        import docker

        return docker.APIClient(
            base_url=self.base_url, version="auto", tls=self.tls_config
        )

    def pull_image(self) -> None:
        """Pull the image specified so it can be built.

        In order for the docker python library to use a base image it must be pulled
        from either the main docker registry or a separate registry that must be set as
        `registry_url` on this class.

        Raises:
            - InterruptedError: if either pulling the image fails
        """
        client = self._get_client()

        output = client.pull(self.base_image, stream=True, decode=True)
        for line in output:
            if line.get("error"):
                raise InterruptedError(line.get("error"))
            if line.get("progress"):
                print(line.get("status"), line.get("progress"), end="\r")
        print("")

    def push_image(self, image_name: str, image_tag: str) -> None:
        """Push this image to a registry

        Args:
            - image_name (str): Name for the image
            - image_tag (str): Tag for the image

        Raises:
            - InterruptedError: if either pushing the image fails
        """
        client = self._get_client()

        self.logger.info("Pushing image to the registry...")

        output = client.push(image_name, tag=image_tag, stream=True, decode=True)
        for line in output:
            if line.get("error"):
                raise InterruptedError(line.get("error"))
            if line.get("progress"):
                print(line.get("status"), line.get("progress"), end="\r")
        print("")

    @staticmethod
    def _parse_generator_output(generator: Iterable) -> None:
        """
        Parses and writes a Docker command's output to stdout
        """
        for item in generator:
            item = item.decode("utf-8")
            for line in item.split("\n"):
                if not line:
                    continue
                parsed = json.loads(line)
                if not isinstance(parsed, dict):
                    continue
                # Parse several possible schemas
                output = (
                    parsed.get("stream")
                    or parsed.get("message")
                    or parsed.get("errorDetail", {}).get("message")
                    or ""
                ).strip("\n")
                if output:
                    print(output)



"""
Prefect Executors are responsible for running tasks in a flow. During
execution of a flow run, a flow's executor will be initialized, used to execute
all tasks in the flow, then shutdown.

Currently, the available executor options are:

- `LocalExecutor`: the default, no frills executor. All tasks are executed in
    a single thread, parallelism is not supported.
- `LocalDaskExecutor`: an executor that runs on `dask` primitives with a
    using either threads or processes.
- `DaskExecutor`: the most feature-rich of the executors, this executor runs
    on `dask.distributed` and has support for distributed execution.

Which executor you choose depends on the performance requirements and
characteristics of your Flow.  See [the executors
docs](/orchestration/flow_config/executors.md) for more information.
"""
from .base import Executor
from .dask import DaskExecutor, LocalDaskExecutor
from .local import LocalExecutor

__all__ = ["DaskExecutor", "Executor", "LocalDaskExecutor", "LocalExecutor"]



"""
Utility functions for interacting with Kubernetes API.
"""
import sys
from typing import Union, Optional

from kubernetes import client, config as kube_config
from kubernetes.config.config_exception import ConfigException

from prefect import config
from prefect.client import Secret
from prefect.utilities.logging import get_logger


K8S_CLIENTS = {
    "job": client.BatchV1Api,
    "pod": client.CoreV1Api,
    "service": client.CoreV1Api,
    "deployment": client.AppsV1Api,
    "secret": client.CoreV1Api,
}

KubernetesClient = Union[client.BatchV1Api, client.CoreV1Api, client.AppsV1Api]

logger = get_logger("kubernetes")


def get_kubernetes_client(
    resource: str, kubernetes_api_key_secret: Optional[str] = "KUBERNETES_API_KEY"
) -> KubernetesClient:
    """
    Utility function for loading kubernetes client object for a given resource.

    It will attempt to connect to a Kubernetes cluster in three steps with
    the first successful connection attempt becoming the mode of communication with a
    cluster.

    1. Attempt to use a Prefect Secret that contains a Kubernetes API Key. If
    `kubernetes_api_key_secret` = `None` then it will attempt the next two connection
    methods. By default the value is `KUBERNETES_API_KEY` so providing `None` acts as
    an override for the remote connection.
    2. Attempt in-cluster connection (will only work when running on a Pod in a cluster)
    3. Attempt out-of-cluster connection using the default location for a kube config file

    In some cases connections to the kubernetes server are dropped after being idle for some time
    (e.g. Azure Firewall drops idle connections after 4 minutes) which would result in
    ReadTimeoutErrors.
    In order to prevent that a periodic keep-alive message can be sent to the server to keep the
    connection open.

    Args:
        - resource (str): the name of the resource to retrieve a client for. Currently
            you can use one of these values: `job`, `pod`, `service`, `deployment`
        - kubernetes_api_key_secret (str, optional): the name of the Prefect Secret
            which stored your Kubernetes API Key; this Secret must be a string and in
            BearerToken format

    Returns:
        - KubernetesClient: an initialized and authenticated Kubernetes Client
    """
    client_type = K8S_CLIENTS[resource]

    if kubernetes_api_key_secret:
        logger.debug("Loading configuration from secret")
        kubernetes_api_key = Secret(kubernetes_api_key_secret).get()
        configuration = client.Configuration()
        configuration.api_key["authorization"] = kubernetes_api_key
        k8s_client = client_type(client.ApiClient(configuration))
    else:
        try:
            logger.debug("Loading incluster configuration")
            kube_config.load_incluster_config()
        except ConfigException as exc:
            logger.warning("{} Using out of cluster configuration option.".format(exc))
            logger.debug("Loading out of cluster configuration")
            kube_config.load_kube_config()

        k8s_client = client_type()

    if config.cloud.agent.kubernetes_keep_alive:
        _keep_alive(client=k8s_client)

    return k8s_client


def _keep_alive(client: KubernetesClient) -> None:
    """
    Setting the keep-alive flags on the kubernetes client object.
    Unfortunately neither the kubernetes library nor the urllib3 library which kubernetes is using
    internally offer the functionality to enable keep-alive messages. Thus the flags are added to
    be used on the underlying sockets.

    Args:
        - client (KubernetesClient): the kubernetes client object on which the keep-alive should be
            enabled
    """
    import socket

    socket_options = [(socket.SOL_SOCKET, socket.SO_KEEPALIVE, 1)]

    if hasattr(socket, "TCP_KEEPINTVL"):
        socket_options.append((socket.IPPROTO_TCP, socket.TCP_KEEPINTVL, 30))

    if hasattr(socket, "TCP_KEEPCNT"):
        socket_options.append((socket.IPPROTO_TCP, socket.TCP_KEEPCNT, 6))

    if hasattr(socket, "TCP_KEEPIDLE"):
        socket_options.append((socket.IPPROTO_TCP, socket.TCP_KEEPIDLE, 6))

    if sys.platform == "darwin":
        # TCP_KEEP_ALIVE not available on socket module in macOS, but defined in
        # https://github.com/apple/darwin-xnu/blob/2ff845c2e033bd0ff64b5b6aa6063a1f8f65aa32/bsd/netinet/tcp.h#L215
        TCP_KEEP_ALIVE = 0x10
        socket_options.append((socket.IPPROTO_TCP, TCP_KEEP_ALIVE, 30))

    client.api_client.rest_client.pool_manager.connection_pool_kw[
        "socket_options"
    ] = socket_options



import os
from typing import Iterable

from prefect.run_configs.base import RunConfig


class LocalRun(RunConfig):
    """Configure a flow-run to run as a Local Process.

    Args:
        - env (dict, optional): Additional environment variables to set for the process
        - working_dir (str, optional): Working directory in which to start the
            process, must already exist. If not provided, will be run in the
            same directory as the agent.
        - labels (Iterable[str], optional): an iterable of labels to apply to this
            run config. Labels are string identifiers used by Prefect Agents
            for selecting valid flow runs when polling for work

    Examples:

    Use the defaults set on the agent:

    ```python
    flow.run_config = LocalRun()
    ```

    Run in a specified working directory:

    ```python
    flow.run_config = LocalRun(working_dir="/path/to/working-directory")
    ```

    Set an environment variable in the flow run process:

    ```python
    flow.run_config = LocalRun(env={"SOME_VAR": "value"})
    ```
    """

    def __init__(
        self,
        *,
        env: dict = None,
        working_dir: str = None,
        labels: Iterable[str] = None,
    ) -> None:
        super().__init__(env=env, labels=labels)
        if working_dir is not None:
            working_dir = os.path.abspath(working_dir)
        self.working_dir = working_dir



"""
Tasks for interacting with monday.com

Note, to authenticate with the Monday API add a Prefect Secret
called '"MONDAY_API_TOKEN"' that stores your Monday API Token.
"""

from prefect.tasks.monday.monday import CreateItem

__all__ = ["CreateItem"]



import time
from typing import Optional

import prefect
from prefect.exceptions import ClientError


def _running_with_backend() -> bool:
    """
    Determine if running in context of a backend. This is always true when running
    using the `CloudTaskRunner`.

    Returns:
        - bool: if `running_with_backend` is set in context
    """
    return bool(prefect.context.get("running_with_backend"))


def _create_task_run_artifact(kind: str, data: dict) -> str:
    client = prefect.Client()
    task_run_id = prefect.context.get("task_run_id")
    # XXX: there's a race condition in the cloud backend for mapped tasks where
    # the task run lookup might fail temporarily. This should last a few
    # seconds max, for now we retry a few times.
    retries = 5
    while True:
        try:
            return client.create_task_run_artifact(
                task_run_id=task_run_id, kind=kind, data=data
            )
        except ClientError as exc:
            # If it's a not found error and we still have retries left, retry
            if "not found" in str(exc).lower() and retries:
                time.sleep(1)
                retries -= 1
                continue
            raise


def create_link_artifact(link: str) -> Optional[str]:
    """
    Create a link artifact

    Args:
        - link (str): the link to post

    Returns:
        - str: the task run artifact ID
    """
    if not _running_with_backend():
        return None
    return _create_task_run_artifact("link", {"link": link})


def update_link_artifact(task_run_artifact_id: str, link: str) -> None:
    """
    Update an existing link artifact. This function will replace the current link
    artifact with the new link provided.

    Args:
        - task_run_artifact_id (str): the ID of an existing task run artifact
        - link (str): the new link to update the artifact with
    """
    if not _running_with_backend():
        return

    client = prefect.Client()
    client.update_task_run_artifact(
        task_run_artifact_id=task_run_artifact_id, data={"link": link}
    )


def create_markdown_artifact(markdown: str) -> Optional[str]:
    """
    Create a markdown artifact

    Args:
        - markdown (str): the markdown to post

    Returns:
        - str: the task run artifact ID
    """
    if not _running_with_backend():
        return None
    return _create_task_run_artifact("markdown", {"markdown": markdown})


def update_markdown_artifact(task_run_artifact_id: str, markdown: str) -> None:
    """
    Update an existing markdown artifact. This function will replace the current markdown
    artifact with the new markdown provided.

    Args:
        - task_run_artifact_id (str): the ID of an existing task run artifact
        - markdown (str): the new markdown to update the artifact with
    """
    if not _running_with_backend():
        return

    client = prefect.Client()
    client.update_task_run_artifact(
        task_run_artifact_id=task_run_artifact_id, data={"markdown": markdown}
    )


def delete_artifact(task_run_artifact_id: str) -> None:
    """
    Delete an existing artifact

    Args:
        - task_run_artifact_id (str): the ID of an existing task run artifact
    """
    if not _running_with_backend():
        return

    client = prefect.Client()
    client.delete_task_run_artifact(task_run_artifact_id=task_run_artifact_id)



"""
Tasks for interacting with SendGrid.
"""
try:
    from prefect.tasks.sendgrid.sendgrid import SendEmail
except ImportError as exc:
    raise ImportError(
        'Using `prefect.tasks.sendgrid` requires Prefect to be installed with the "sendgrid" extra.'
    ) from exc

__all__ = ["SendEmail"]



"""
This module contains a collection of tasks for interacting with Exasol databases via
the pyexasol library.
"""
try:
    from prefect.tasks.exasol.exasol import (
        ExasolFetch,
        ExasolExecute,
        ExasolImportFromIterable,
        ExasolExportToFile,
    )
except ImportError as exc:
    raise ImportError(
        'Using `prefect.tasks.exasol` requires Prefect to be installed with the "exasol" extra.'
    ) from exc

__all__ = [
    "ExasolExecute",
    "ExasolExportToFile",
    "ExasolFetch",
    "ExasolImportFromIterable",
]



from prefect.tasks.control_flow.conditional import ifelse, switch, merge
from prefect.tasks.control_flow.filter import FilterTask
from prefect.tasks.control_flow.case import case

__all__ = ["FilterTask", "case", "ifelse", "merge", "switch"]



import inspect
import json
import pickle
from decimal import Decimal
from functools import partial, wraps
from typing import Any, Callable, Iterable, Type, Optional

from prefect.client import Secret

from toloka.client import TolokaClient, add_headers
import toloka.client as _toloka_client_lib  # To avoid `structure` pickling errors.


def with_updated_signature(
    func: Callable,
    wrapper: Callable,
    add_wrapper_args: Iterable[str] = (),
    remove_func_args: Iterable[str] = (),
) -> Callable:
    """
    Prefect tasks relate on function signature,
    so we need to change it since we use some args-changing wrappers.

    Args:
        - func (Callable): Initial function to get signature from.
        - wrapper (Callable): Wrapper to take body from.
        - add_wrapper_args (Iterable, optional): Wrapper's args to put in the final signature.
        - remove_func_args (Iterable, optional): Func's args to remove from the final signature.

    Returns:
        - Callable: initial wrapper with modified signature.
    """

    remove_func_args = set(remove_func_args)

    signature_func = inspect.signature(func)
    parameters_keep = [
        param
        for param in signature_func.parameters.values()
        if param.name not in remove_func_args
        and param.kind != inspect.Parameter.VAR_KEYWORD
    ]

    parameters_wrapper = inspect.signature(wrapper).parameters
    parameters_add = [parameters_wrapper[arg_name] for arg_name in add_wrapper_args]

    last_param = next(reversed(signature_func.parameters.values()), inspect.Parameter)
    if last_param.kind == inspect.Parameter.VAR_KEYWORD:
        parameters_add.append(last_param)

    res = wraps(func)(wrapper)
    res.__special_wrapped__ = res.__wrapped__
    del res.__wrapped__
    res.__signature__ = signature_func.replace(
        parameters=parameters_keep + parameters_add
    )

    return res


def with_logger(func: Callable) -> Callable:
    """
    Decorator that allows function to use Prefect logger instance.
    Enrich function signature with keyword-only argument "logger".
    Use it to write logs from Prefect task.

    Args:
        - func (Callable): Function that takes `logger` argument.

    Returns:
        - Callable: a wrapper with Prefect logger passed.

    Example:
        >>> @task
        ... def some_task():
        ...     return 123
        ...
        >>> @task
        ... @with_logger
        ... def task_with_logger(arg, logger):
        ...     logger.warning('arg value: %s', arg)
        ...
        >>> with Flow('some-flow') as flow:
        ...     res = some_task()
        ...     task_with_logger(res)
        ...
    """

    def _wrapper(*args, **kwargs) -> Any:
        import prefect

        return partial(func, logger=prefect.context.get("logger"))(*args, **kwargs)

    return with_updated_signature(func, _wrapper, remove_func_args={"logger"})


_json_loads = partial(json.loads, parse_float=Decimal)


def structure_from_conf(obj: Any, cl: Type) -> object:
    if isinstance(obj, cl):
        return obj
    if isinstance(obj, bytes):
        try:
            return pickle.loads(obj)
        except Exception:
            pass
        obj = obj.decode()
    if isinstance(obj, str):
        obj = _json_loads(obj)
    return _toloka_client_lib.structure(obj, cl)


def extract_id(obj: Any, cl: Type) -> str:
    if isinstance(obj, str):
        try:
            obj = _json_loads(obj)
        except Exception:
            return obj
        if isinstance(obj, int):
            return str(obj)
    res = structure_from_conf(obj, cl).id
    if res is None:
        raise ValueError(f"Got id=None from: {obj}")
    return res


DEFAULT_TOLOKA_SECRET_NAME = "TOLOKA_TOKEN"
DEFAULT_TOLOKA_ENV = "PRODUCTION"


def with_toloka_client(func: Callable) -> Callable:
    """
    Decorator that allows function to pass `token`, `secret_name` and `env` args
    and operate with `toloka_client` instance.

    Args:
        - func (Callable): Function that operate with `toloka_client` argument.

    Returns:
        - Callable: The wrapper, which takes optional `token`, `secret_name` and `env` arguments
            and operates with default `toloka_token` if they are not passed.

    Example:
        >>> @with_toloka_client
        ... def some_func(toloka_client: TolokaClient):
        ...     toloka_client.create_project(...)
        ...
        >>> some_func()  # Use default toloka_client created using TOLOKA_TOKEN secret.
        >>> some_func(secret_name="OTHER_ACCOUNT_SECRET")  # Allow to pass other secret name.
        >>> some_func(token=PrefectSecret("TOLOKA_TOKEN")) # Provide secret directly.
        ...
    """

    def _wrapper(
        *args,
        token: Optional[str] = None,
        secret_name: str = DEFAULT_TOLOKA_SECRET_NAME,
        env: str = DEFAULT_TOLOKA_ENV,
        **kwargs,
    ) -> Any:
        token = token or Secret(secret_name).get()
        toloka_client = TolokaClient(token, env)
        return partial(add_headers("prefect")(func), toloka_client=toloka_client)(
            *args, **kwargs
        )

    return with_updated_signature(
        func,
        _wrapper,
        remove_func_args=("toloka_client",),
        add_wrapper_args=("token", "secret_name", "env"),
    )



"""
Utility functions for interacting with and configuring logging.  The main entrypoint for
retrieving loggers for customization is the `get_logger` utility.

Note that Prefect Tasks come equipped with their own loggers.  These can be accessed via:
    - `self.logger` if implementing a Task class
    - `prefect.context.get("logger")` if using the `task` decorator

When running locally, log levels and message formatting are set via your Prefect configuration file.
"""
import atexit
import logging
import sys
import threading
import time
import warnings
import json
from queue import Empty, Queue
from typing import Any, List, Optional, Generator, Union
from contextlib import contextmanager
from uuid import uuid4

import pendulum

import prefect
from prefect.utilities.context import context

_original_log_record_factory = logging.getLogRecordFactory()

PREFECT_LOG_RECORD_ATTRIBUTES = (
    "flow_name",
    "flow_run_id",
    "task_name",
    "task_slug",
    "task_run_id",
)

MAX_LOG_SIZE = 1_000_000  # 1 MB - max size of a single log record
MAX_BATCH_LOG_SIZE = 4_000_000  # 4 MB - max total batch size for log records

LOG_OVERHEAD = len(
    # Calculate the expected overhead from a log record by serializing a record with
    # metadata
    json.dumps(
        {
            "flow_run_id": uuid4().hex,
            "task_run_id": uuid4().hex,
            "timestamp": pendulum.from_timestamp(time.time()).isoformat(),
            "name": "X" * 100,
            "level": "WARNING",
            "message": "",
        }
    )
)


def getlogsize(log: dict) -> int:
    return len(log.get("message", "")) + LOG_OVERHEAD


class LogManager:
    """A global log manager for managing all logs to be sent to Prefect"""

    def __init__(self) -> None:
        self.queue = Queue()  # type: Queue[dict]
        self.pending_logs = []  # type: List[dict]
        self.thread = None  # type: Optional[threading.Thread]
        self.client = None  # type: Optional[prefect.Client]
        self.pending_length = 0
        self._stopped = threading.Event()

    def ensure_started(self) -> None:
        """Ensure the log manager is started"""
        if self.thread is None:
            self.client = prefect.Client()
            self.logging_period = context.config.cloud.logging_heartbeat
            self.thread = threading.Thread(
                target=self._write_logs_loop,
                name="prefect-log-manager",
                daemon=True,
            )
            self.thread.start()
            atexit.register(self._on_shutdown)

    def _on_shutdown(self) -> None:
        """Called via atexit, flushes all logs and stops the background thread"""
        # Sometimes a signal can hit the process at shutdown multiple times,
        # interrupting an active shutdown hook. To give a better chance of the
        # shutdown hook succeeding, we retry a few times, ignoring extra
        # `SystemExit` exceptions raised here. Note that this won't prevent
        # shutdown (the interpreter is already shutting down regardless).
        for _ in range(3):
            try:
                self.stop()
                return
            except SystemExit:
                pass

    def stop(self) -> None:
        """Flush all logs and stop the background thread"""
        if self.thread is not None:
            self._stopped.set()
            self.thread.join()
            self._write_logs()
            self.thread = None
            self.client = None

    def _write_logs_loop(self) -> None:
        """Runs in a background thread, uploads logs periodically in a loop"""
        while not self._stopped.wait(self.logging_period):
            self._write_logs()

    def _write_logs(self) -> None:
        """Upload logs in batches until the queue is empty"""
        assert self.client is not None  # mypy

        # Read all logs from the queue into the `pending_logs` list. This
        # is stored on the manager to ensure that logs aren't dropped in
        # the case of an upload error, and will be retried later. This
        # could be due to an api error, or due to a shutdown signal
        # interrupting a log upload.
        #
        # We batch uploads with a max total length to prevent uploading too
        # large a payload at once. This call will continue to loop until
        # the queue is empty or an error occurs on upload (usually only one
        # iteration is sufficient)
        max_batch_length = max(MAX_BATCH_LOG_SIZE - MAX_LOG_SIZE, MAX_LOG_SIZE)
        cont = True
        while cont:
            try:
                while self.pending_length < max_batch_length:
                    log = self.queue.get_nowait()
                    self.pending_length += getlogsize(log)
                    self.pending_logs.append(log)

            except Empty:
                cont = False

            if self.pending_logs:
                try:
                    self.client.write_run_logs(self.pending_logs)
                    self.pending_logs = []
                    self.pending_length = 0
                except Exception as exc:
                    # An error occurred on upload, warn and exit the loop (will
                    # retry later)
                    warnings.warn(
                        f"Failed to write logs with error: {exc!r}, "
                        f"Pending log length: {self.pending_length:,}, "
                        f"Max batch log length: {MAX_BATCH_LOG_SIZE:,}, "
                        f"Queue size: {self.queue.qsize():,}"
                    )
                    cont = False

    def enqueue(self, message: dict) -> None:
        """Enqueue a new log message to be uploaded.

        Args:
            - message (dict): a log message to upload.
        """
        self.ensure_started()
        self.queue.put(message)


class CloudHandler(logging.Handler):
    """A handler for sending logs to the prefect API"""

    def emit(self, record: logging.LogRecord) -> None:  # type: ignore
        """Emit a new log"""
        # if we shouldn't log to cloud, don't emit
        if not context.config.cloud.send_flow_run_logs:
            return

        # if its not during a backend flow run, don't emit
        if not context.get("running_with_backend"):
            return

        # ensures emitted logs respect configured logging level
        config_level = getattr(logging, context.config.logging.level, logging.INFO)

        if record.levelno < config_level:
            return

        msg = self.format(record)

        log = {
            "flow_run_id": context.get("flow_run_id"),
            "task_run_id": context.get("task_run_id"),
            "timestamp": pendulum.from_timestamp(
                getattr(record, "created", None) or time.time()
            ).isoformat(),
            "name": getattr(record, "name", None),
            "level": getattr(record, "levelname", None),
            "message": msg,
        }

        log_size = getlogsize(log)
        if log_size > MAX_LOG_SIZE:
            warnings.warn(
                "Received a log of size %d bytes, exceeding the limit of %d. "
                "The message will be truncated." % (log_size, MAX_LOG_SIZE)
            )
            log["message"] = msg[: MAX_LOG_SIZE - LOG_OVERHEAD]

        LOG_MANAGER.enqueue(log)


def _log_record_context_injector(*args: Any, **kwargs: Any) -> logging.LogRecord:
    """
    A custom logger LogRecord Factory that injects selected context parameters into newly
    created logs.

    Args:
        - *args: arguments to pass to the original LogRecord Factory
        - **kwargs: keyword arguments to pass to the original LogRecord Factory

    Returns:
        - logging.LogRecord: the newly created LogRecord
    """
    record = _original_log_record_factory(*args, **kwargs)

    additional_attrs = context.config.logging.get("log_attributes", [])

    for attr in PREFECT_LOG_RECORD_ATTRIBUTES + tuple(additional_attrs):
        value = context.get(attr, None)
        if value or attr in additional_attrs:
            setattr(record, attr, value)

    return record


def _create_logger(name: str) -> logging.Logger:
    """
    Creates a logger with a `StreamHandler` that has level and formatting
    set from `prefect.config`.

    Args:
        - name (str): Name to use for logger.

    Returns:
        - logging.Logger: a configured logging object
    """
    logging.setLogRecordFactory(_log_record_context_injector)

    logger = logging.getLogger(name)

    # Set the format from the config for stdout
    formatter = logging.Formatter(
        context.config.logging.format, context.config.logging.datefmt
    )
    stream_handler = logging.StreamHandler(sys.stdout)
    stream_handler.setFormatter(formatter)
    logger.addHandler(stream_handler)

    # Set the level
    logger.setLevel(context.config.logging.level)
    logger.addHandler(CloudHandler())

    return logger


def configure_logging(testing: bool = False) -> logging.Logger:
    """
    Creates a "prefect" root logger with a `StreamHandler` that has level and formatting
    set from `prefect.config`.

    Args:
        - testing (bool, optional): a boolean specifying whether this configuration
            is for testing purposes only; this helps us isolate any global state during testing
            by configuring a "prefect-test-logger" instead of the standard "prefect" logger

    Returns:
        - logging.Logger: a configured logging object
    """
    name = "prefect-test-logger" if testing else "prefect"

    return _create_logger(name)


context.logger = prefect_logger = configure_logging()


def configure_extra_loggers() -> None:
    """
    Creates a "Prefect" configured logger for all strings in extra_loggers config list.
    The logging.extra_loggers config defaults to an empty list.
    """
    loggers = context.config.logging.get("extra_loggers", [])
    for l in loggers:
        _create_logger(l)


configure_extra_loggers()


def create_diagnostic_logger(name: str) -> logging.Logger:
    """
    Create a logger that does not use the `CloudHandler` but preserves all other
    Prefect logging configuration.  For diagnostic / debugging / internal use only.
    """
    logger = _create_logger(name)
    logger.handlers = [h for h in logger.handlers if not isinstance(h, CloudHandler)]
    return logger


def get_logger(name: str = None) -> logging.Logger:
    """
    Returns a "prefect" logger.

    Args:
        - name (str): if `None`, the root Prefect logger is returned. If provided, a child
            logger of the name `"prefect.{name}"` is returned. The child logger inherits
            the root logger's settings.

    Returns:
        - logging.Logger: a configured logging object with the appropriate name
    """

    if name is None:
        return prefect_logger
    else:
        return prefect_logger.getChild(name)


@contextmanager
def temporary_logger_config(
    level: Union[int, str] = None,
    stream_fmt: str = None,
    stream_datefmt: str = None,
) -> Generator[logging.Logger, None, None]:
    """
    Set a temporary config for the `prefect` logger. The formatting can be updated
    for `StreamHandlers` but will not update the CloudHandler (or any others)

    This function is only intended to be used internally.

    Args:
        - level (optional): The log level to use
        - stream_fmt (optional): The log message format to set for stream handlers
        - stream_datefmt (optional): The log date format to set for stream handlers

    Yields:
        The configured logger; also affects any `prefect` loggers in use
    """
    logger = get_logger()

    # This is the only key that seems likely to be adjusted after the config has been
    # loaded
    previous_log_level = logger.level

    overrides = {
        "logging.level": level if level else None,
        "logging.format": stream_fmt,
        "logging.datefmt": stream_datefmt,
    }
    # Drop empty values to retain existing settings
    overrides = {key: value for key, value in overrides.items() if value}

    try:
        with prefect.utilities.configuration.set_temporary_config(overrides):
            logger.handlers.clear()
            logger = configure_logging()
            yield logger
    finally:
        # Restore the old logger
        logger = get_logger()
        logger.handlers.clear()
        logger = configure_logging()
        logger.setLevel(previous_log_level)


LOG_MANAGER = LogManager()


class RedirectToLog:
    """
    Custom redirect of stdout messages to logs

    Args:
        - logger (logging.Logger, optional): an optional logger to redirect stdout. If
            not provided a logger names `stdout` will be created.
    """

    def __init__(self, logger: logging.Logger = None) -> None:
        self.stdout_logger = logger or get_logger("stdout")

    def write(self, s: str) -> None:
        """
        Write message from stdout to a prefect logger.
        Note: blank newlines will not be logged.

        Args:
            s (str): the message from stdout to be logged
        """
        if not isinstance(s, str):
            # stdout is expecting str
            raise TypeError(f"string argument expected, got {type(s)}")

        if s.strip():
            self.stdout_logger.info(s)

    def flush(self) -> None:
        """
        Implemented flush operation for logger handler
        """
        for handler in self.stdout_logger.handlers:
            handler.flush()



"""
This module contains a collection of tasks to interact with Neo4j.
"""

try:
    from prefect.tasks.neo4j.neo4j_tasks import Neo4jRunCypherQueryTask
except ImportError as err:
    raise ImportError(
        'Using `prefect.tasks.neo4j` requires Prefect to be installed with the "neo4j" extra.'
    ) from err

__all__ = ["Neo4jRunCypherQueryTask"]



from .base import RunConfig, UniversalRun
from .kubernetes import KubernetesRun
from .local import LocalRun
from .docker import DockerRun
from .ecs import ECSRun
from .vertex import VertexRun

__all__ = [
    "DockerRun",
    "ECSRun",
    "KubernetesRun",
    "LocalRun",
    "RunConfig",
    "UniversalRun",
    "VertexRun",
]



from typing import Dict, List, Union, Optional

import azureml.core.dataset
from azureml.core.datastore import Datastore
from azureml.data import DataType, TabularDataset
from azureml.data.dataset_type_definitions import PromoteHeadersBehavior

from prefect import Task
from prefect.utilities.tasks import defaults_from_attrs


class DatasetCreateFromDelimitedFiles(Task):
    """
    Task for creating a TabularDataset from delimited files for use in a Azure Machine Learning
    service Workspace. The files should exist in a Datastore.  Note that all initialization
    arguments can optionally be provided or overwritten at runtime.

    Args:
        - dataset_name (str, optional): The name of the Dataset in the Workspace
        - datastore (azureml.core.datastore.Datastore, optional): The Datastore which holds the
            files.
        - path (Union[str, List[str]], optional): The path to the delimited files in the
            Datastore.
        - dataset_description (str, optional): Description of the Dataset.
        - dataset_tags (str, optional): Tags to associate with the Dataset.
        - include_path (bool, optional): Boolean to keep path information as column in the
            dataset.
        - infer_column_types (bool, optional): Boolean to infer column data types.
        - set_column_types (Dict[str, azureml.data.DataType], optional): A dictionary to set
            column data type, where key is column name and value is a `azureml.data.DataType`.
        - fine_grain_timestamp (str, optional): The name of column as fine grain timestamp.
        - coarse_grain_timestamp (str, optional): The name of column coarse grain timestamp.
        - separator (str, optional): The separator used to split columns.
        - header (azureml.data.dataset_type_definitions.PromoteHeadersBehavior, optional):
            Controls how column headers are promoted when reading from files. Defaults to
            assume that all files have the same header.
        - partition_format (str, optional): Specify the partition format of path. Defaults to
            None. The partition information of each path will be extracted into columns based
            on the specified format. Format part `{column_name}` creates string column, and
            `{column_name:yyyy/MM/dd/HH/mm/ss}` creates datetime column, where `yyyy`, `MM`,
            `dd`, `HH`, `mm` and `ss` are used to extrat year, month, day, hour, minute and
            second for the datetime type. The format should start from the position of first
            partition key until the end of file path. For example, given the path
            `../Germany/2019/01/01/data.csv` where the partition is by country and time,
            `partition_format="/{Country}/{PartitionDate:yyyy/MM/dd}/data.csv"` creates string
            column `Country` with value `Germany` and datetime column `PartitionDate` with
            value `2019-01-01`.
        - create_new_version (bool, optional): Boolean to register the dataset as a new version
            under the specified name.
        - **kwargs (dict, optional): additional keyword arguments to pass to the Task
            constructor
    """

    def __init__(
        self,
        dataset_name: str = None,
        datastore: Datastore = None,
        path: Union[str, List[str]] = None,
        dataset_description: str = "",
        dataset_tags: Optional[Dict[str, str]] = None,
        include_path: bool = False,
        infer_column_types: bool = True,
        set_column_types: Optional[Dict[str, DataType]] = None,
        fine_grain_timestamp: str = None,
        coarse_grain_timestamp: str = None,
        separator: str = ",",
        header: PromoteHeadersBehavior = PromoteHeadersBehavior.ALL_FILES_HAVE_SAME_HEADERS,
        partition_format: str = None,
        create_new_version: bool = False,
        **kwargs
    ) -> None:
        self.dataset_name = dataset_name
        self.datastore = datastore
        self.path = path
        self.dataset_description = dataset_description
        self.dataset_tags = dataset_tags or dict()
        self.include_path = include_path
        self.infer_column_types = infer_column_types
        self.set_column_types = set_column_types
        self.fine_grain_timestamp = fine_grain_timestamp
        self.coarse_grain_timestamp = coarse_grain_timestamp
        self.separator = separator
        self.header = header
        self.partition_format = partition_format
        self.create_new_version = create_new_version

        super().__init__(**kwargs)

    @defaults_from_attrs(
        "dataset_name",
        "datastore",
        "path",
        "dataset_description",
        "dataset_tags",
        "include_path",
        "infer_column_types",
        "set_column_types",
        "fine_grain_timestamp",
        "coarse_grain_timestamp",
        "separator",
        "header",
        "partition_format",
        "create_new_version",
    )
    def run(
        self,
        dataset_name: str = None,
        datastore: Datastore = None,
        path: Union[str, List[str]] = None,
        dataset_description: str = "",
        dataset_tags: Optional[Dict[str, str]] = None,
        include_path: bool = False,
        infer_column_types: bool = True,
        set_column_types: Optional[Dict[str, DataType]] = None,
        fine_grain_timestamp: str = None,
        coarse_grain_timestamp: str = None,
        separator: str = ",",
        header: PromoteHeadersBehavior = PromoteHeadersBehavior.ALL_FILES_HAVE_SAME_HEADERS,
        partition_format: str = None,
        create_new_version: bool = False,
    ) -> TabularDataset:
        """
        Task run method.

        Args:
            - dataset_name (str, optional): The name of the Dataset in the Workspace
            - datastore (azureml.core.datastore.Datastore, optional): The Datastore which holds
                the files.
            - path (Union[str, List[str]], optional): The path to the delimited files in the
                Datastore.
            - dataset_description (str, optional): Description of the Dataset.
            - dataset_tags (str, optional): Tags to associate with the Dataset.
            - include_path (bool, optional): Boolean to keep path information as column in the
                dataset.
            - infer_column_types (bool, optional): Boolean to infer column data types.
            - set_column_types (Dict[str, azureml.data.DataType], optional): A dictionary to
                set column data type, where key is column name and value is a
                `azureml.data.DataType`.
            - fine_grain_timestamp (str, optional): The name of column as fine grain timestamp.
            - coarse_grain_timestamp (str, optional): The name of column coarse grain timestamp.
            - separator (str, optional): The separator used to split columns.
            - header (azureml.data.dataset_type_definitions.PromoteHeadersBehavior, optional):
                Controls how column headers are promoted when reading from files. Defaults to
                assume that all files have the same header.
            - partition_format (str, optional): Specify the partition format of path.
            - create_new_version (bool, optional): Boolean to register the dataset as a new
                version under the specified name.

        Returns:
            - azureml.data.TabularDataset: the created TabularDataset

        """
        if dataset_name is None:
            raise ValueError("A dataset_name must be provided.")

        if path is None:
            raise ValueError("A path must be provided.")

        if datastore is None:
            raise ValueError("A path must be provided.")

        if not isinstance(path, list):
            path = [path]

        dataset_tags = dataset_tags or dict()

        dataset = azureml.core.dataset.Dataset.Tabular.from_delimited_files(
            path=[(datastore, path_item) for path_item in path],
            include_path=include_path,
            infer_column_types=infer_column_types,
            set_column_types=set_column_types,
            separator=separator,
            header=header,
            partition_format=partition_format,
        )

        dataset = dataset.with_timestamp_columns(
            fine_grain_timestamp=fine_grain_timestamp,
            coarse_grain_timestamp=coarse_grain_timestamp,
            validate=True,
        )

        dataset = dataset.register(
            workspace=datastore.workspace,
            name=dataset_name,
            description=dataset_description,
            tags=dataset_tags,
            create_new_version=create_new_version,
        )

        return dataset


class DatasetCreateFromParquetFiles(Task):
    """
    Task for creating a TabularDataset from Parquet files for use in a Azure Machine Learning
    service Workspace. The files should exist in a Datastore.  Note that all initialization
    arguments can optionally be provided or overwritten at runtime.

    Args:
        - dataset_name (str, optional): The name of the Dataset in the Workspace
        - datastore (azureml.core.datastore.Datastore, optional): The Datastore which holds the
            files.
        - path (Union[str, List[str]], optional): The path to the delimited files in the Datastore.
        - dataset_description (str, optional): Description of the Dataset.
        - dataset_tags (str, optional): Tags to associate with the Dataset.
        - include_path (bool, optional): Boolean to keep path information as column in the dataset.
        - set_column_types (Dict[str, azureml.data.DataType], optional): A dictionary to set
            column data type, where key is column name and value is a `azureml.data.DataType`.
        - fine_grain_timestamp (str, optional): The name of column as fine grain timestamp.
        - coarse_grain_timestamp (str, optional): The name of column coarse grain timestamp.
        - partition_format (str, optional): Specify the partition format of path. Defaults to
            None. The partition information of each path will be extracted into columns based
            on the specified format. Format part `{column_name}` creates string column, and
            `{column_name:yyyy/MM/dd/HH/mm/ss}` creates datetime column, where `yyyy`, `MM`,
            `dd`, `HH`, `mm` and `ss` are used to extrat year, month, day, hour, minute and
            second for the datetime type. The format should start from the position of first
            partition key until the end of file path. For example, given the path
            `../Germany/2019/01/01/data.csv` where the partition is by country and time,
            `partition_format="/{Country}/{PartitionDate:yyyy/MM/dd}/data.csv"` creates string
            column `Country` with value `Germany` and datetime column `PartitionDate` with
            value `2019-01-01`.
        - create_new_version (bool, optional): Boolean to register the dataset as a new version
            under the specified name.
        - **kwargs (dict, optional): additional keyword arguments to pass to the
            Task constructor
    """

    def __init__(
        self,
        dataset_name: str = None,
        datastore: Datastore = None,
        path: Union[str, List[str]] = None,
        dataset_description: str = "",
        dataset_tags: Optional[Dict[str, str]] = None,
        include_path: bool = False,
        set_column_types: Optional[Dict[str, DataType]] = None,
        fine_grain_timestamp: str = None,
        coarse_grain_timestamp: str = None,
        partition_format: str = None,
        create_new_version: bool = False,
        **kwargs
    ) -> None:
        self.dataset_name = dataset_name
        self.datastore = datastore
        self.path = path
        self.dataset_description = dataset_description
        self.dataset_tags = dataset_tags or dict()
        self.include_path = include_path
        self.set_column_types = set_column_types
        self.fine_grain_timestamp = fine_grain_timestamp
        self.coarse_grain_timestamp = coarse_grain_timestamp
        self.partition_format = partition_format
        self.create_new_version = create_new_version

        super().__init__(**kwargs)

    @defaults_from_attrs(
        "dataset_name",
        "datastore",
        "path",
        "dataset_description",
        "dataset_tags",
        "include_path",
        "set_column_types",
        "fine_grain_timestamp",
        "coarse_grain_timestamp",
        "partition_format",
        "create_new_version",
    )
    def run(
        self,
        dataset_name: str = None,
        datastore: Datastore = None,
        path: Union[str, List[str]] = None,
        dataset_description: str = "",
        dataset_tags: Optional[Dict[str, str]] = None,
        include_path: bool = False,
        set_column_types: Optional[Dict[str, DataType]] = None,
        fine_grain_timestamp: str = None,
        coarse_grain_timestamp: str = None,
        partition_format: str = None,
        create_new_version: bool = None,
    ) -> TabularDataset:
        """
        Task run method.

        Args:
            - dataset_name (str, optional): The name of the Dataset in the Workspace
            - datastore (azureml.core.datastore.Datastore, optional): The Datastore which holds
                the files.
            - path (Union[str, List[str]], optional): The path to the delimited files in the
                Datastore.
            - dataset_description (str, optional): Description of the Dataset.
            - dataset_tags (str, optional): Tags to associate with the Dataset.
            - include_path (bool, optional): Boolean to keep path information as column in the
                dataset.
            - set_column_types (Dict[str, azureml.data.DataType], optional): A dictionary to
                set column data type, where key is column name and value is a
                `azureml.data.DataType`.
            - fine_grain_timestamp (str, optional): The name of column as fine grain timestamp.
            - coarse_grain_timestamp (str, optional): The name of column coarse grain timestamp.
            - partition_format (str, optional): Specify the partition format of path.
            - create_new_version (bool, optional): Boolean to register the dataset as a new
                version under the specified name.

        Returns:
            - azureml.data.TabularDataset: the created TabularDataset.
        """
        if dataset_name is None:
            raise ValueError("A dataset_name must be provided.")

        if path is None:
            raise ValueError("A path must be provided.")

        if datastore is None:
            raise ValueError("A path must be provided.")

        if not isinstance(path, list):
            path = [path]

        dataset_tags = dataset_tags or dict()

        dataset = azureml.core.dataset.Dataset.Tabular.from_parquet_files(
            path=[(datastore, path_item) for path_item in path],
            include_path=include_path,
            set_column_types=set_column_types,
            partition_format=partition_format,
        )

        dataset = dataset.with_timestamp_columns(
            fine_grain_timestamp=fine_grain_timestamp,
            coarse_grain_timestamp=coarse_grain_timestamp,
            validate=True,
        )

        dataset = dataset.register(
            workspace=datastore.workspace,
            name=dataset_name,
            description=dataset_description,
            tags=dataset_tags,
            create_new_version=create_new_version,
        )

        return dataset


class DatasetCreateFromFiles(Task):
    """
    Task for creating a FileDataset from files for use in a Azure Machine Learning service
    Workspace. The files should exist in a Datastore.  Note that all initialization arguments
    can optionally be provided or overwritten at runtime.

    Args:
        - dataset_name (str, optional): The name of the Dataset in the Workspace
        - datastore (azureml.core.datastore.Datastore, optional): The Datastore which holds the
            files.
        - path (Union[str, List[str]], optional): The path to the delimited files in the Datastore.
        - dataset_description (str, optional): Description of the Dataset.
        - dataset_tags (str, optional): Tags to associate with the Dataset.
        - create_new_version (bool, optional): Boolean to register the dataset as a new version
            under the specified name.
        - **kwargs (dict, optional): additional keyword arguments to pass to the
            Task constructor
    """

    def __init__(
        self,
        dataset_name: str = None,
        datastore: Datastore = None,
        path: Union[str, List[str]] = None,
        dataset_description: str = "",
        dataset_tags: Optional[Dict[str, str]] = None,
        create_new_version: bool = False,
        **kwargs
    ) -> None:
        self.dataset_name = dataset_name
        self.datastore = datastore
        self.path = path
        self.dataset_description = dataset_description
        self.dataset_tags = dataset_tags or dict()
        self.create_new_version = create_new_version

        super().__init__(**kwargs)

    @defaults_from_attrs(
        "dataset_name",
        "datastore",
        "path",
        "dataset_description",
        "dataset_tags",
        "create_new_version",
    )
    def run(
        self,
        dataset_name: str = None,
        datastore: Datastore = None,
        path: Union[str, List[str]] = None,
        dataset_description: str = "",
        dataset_tags: Optional[Dict[str, str]] = None,
        create_new_version: bool = False,
    ) -> azureml.data.FileDataset:
        """
        Task run method.

        Args:
            - dataset_name (str, optional): The name of the Dataset in the Workspace
            - datastore (azureml.core.datastore.Datastore, optional): The Datastore which holds
                the files.
            - path (Union[str, List[str]], optional): The path to the delimited files in the
                Datastore.
            - dataset_description (str, optional): Description of the Dataset.
            - dataset_tags (str, optional): Tags to associate with the Dataset.
            - create_new_version (bool, optional): Boolean to register the dataset as a new
                version under the specified name.

        Returns:
            - azureml.data.FileDataset: the created FileDataset
        """
        if dataset_name is None:
            raise ValueError("A dataset_name must be provided.")

        if path is None:
            raise ValueError("A path must be provided.")

        if datastore is None:
            raise ValueError("A path must be provided.")

        if not isinstance(path, list):
            path = [path]

        dataset_tags = dataset_tags or dict()

        dataset = azureml.core.dataset.Dataset.File.from_files(
            path=[(datastore, path_item) for path_item in path]
        )

        dataset = dataset.register(
            workspace=datastore.workspace,
            name=dataset_name,
            description=dataset_description,
            tags=dataset_tags,
            create_new_version=create_new_version,
        )

        return dataset
